"""Metadata and database operations client functionality."""

from typing import Any

from ... import fragments
from ...types import (
    AnonymiseDatabaseInput,
    AutoTagMetadataInput,
    AutoTagMetadataOptions,
    BackupDatabaseInput,
    CleanGeneratedInput,
    CleanMetadataInput,
    ConfigDefaultSettingsResult,
    ConfigResult,
    ExportObjectsInput,
    GenerateMetadataInput,
    GenerateMetadataOptions,
    IdentifyMetadataInput,
    ImportObjectsInput,
    MigrateBlobsInput,
    MigrateInput,
    MigrateSceneScreenshotsInput,
    ScanMetadataInput,
    ScanMetadataOptions,
    SetupInput,
)
from ...types.unset import UnsetType, is_set
from ..protocols import StashClientProtocol


class MetadataClientMixin(StashClientProtocol):
    """Mixin for metadata and database operation methods."""

    def _dump_metadata_options(
        self, options: GenerateMetadataOptions
    ) -> dict[str, Any]:
        exclude_fields = {
            field_name
            for field_name in options.__class__.model_fields
            if isinstance(getattr(options, field_name, None), UnsetType)
        }
        return options.model_dump(mode="json", exclude=exclude_fields)

    async def metadata_generate(
        self,
        options: GenerateMetadataOptions | dict[str, Any] | None = None,
        input_data: GenerateMetadataInput | dict[str, Any] | None = None,
    ) -> str:
        """Generate metadata.

        Args:
            options: GenerateMetadataOptions object or dictionary of what to generate:
                - covers: bool - Generate covers
                - sprites: bool - Generate sprites
                - previews: bool - Generate previews
                - imagePreviews: bool - Generate image previews
                - previewOptions: GeneratePreviewOptionsInput
                    - previewSegments: int - Number of segments in a preview file
                    - previewSegmentDuration: float - Duration of each segment in seconds
                    - previewExcludeStart: str - Duration to exclude from start
                    - previewExcludeEnd: str - Duration to exclude from end
                    - previewPreset: PreviewPreset - Preset when generating preview
                - markers: bool - Generate markers
                - markerImagePreviews: bool - Generate marker image previews
                - markerScreenshots: bool - Generate marker screenshots
                - transcodes: bool - Generate transcodes
                - forceTranscodes: bool - Generate transcodes even if not required
                - phashes: bool - Generate phashes
                - interactiveHeatmapsSpeeds: bool - Generate interactive heatmaps speeds
                - imageThumbnails: bool - Generate image thumbnails
                - clipPreviews: bool - Generate clip previews
            input_data: Optional GenerateMetadataInput object or dictionary to specify what to process:
                - sceneIDs: list[str] - List of scene IDs to generate for (default: all)
                - markerIDs: list[str] - List of marker IDs to generate for (default: all)
                - overwrite: bool - Overwrite existing media (default: False)

        Returns:
            Job ID for the generation task

        Raises:
            ValueError: If the input data is invalid
            gql.TransportError: If the request fails
        """
        # Convert GenerateMetadataOptions to dict if needed
        options_dict = {}
        if options is not None:
            if isinstance(options, GenerateMetadataOptions):
                options_dict = self._dump_metadata_options(options)
            else:
                if not isinstance(options, dict):
                    raise TypeError(
                        "options must be GenerateMetadataOptions or dict, "
                        f"got {type(options).__name__}"
                    )
                validated_options = GenerateMetadataOptions(**options)
                options_dict = self._dump_metadata_options(validated_options)

        # Convert GenerateMetadataInput to dict if needed
        input_dict = {}
        if input_data is not None:
            if isinstance(input_data, GenerateMetadataInput):
                input_dict = input_data.to_graphql()
            else:
                if not isinstance(input_data, dict):
                    raise TypeError(
                        "input_data must be GenerateMetadataInput or dict, "
                        f"got {type(input_data).__name__}"
                    )
                validated_input = GenerateMetadataInput(**input_data)
                input_dict = validated_input.to_graphql()

        try:
            # Combine options and input data
            variables = {"input": {**options_dict, **input_dict}}

            # Execute mutation with combined input
            result = await self.execute(fragments.METADATA_GENERATE_MUTATION, variables)

            job_id = result.get("metadataGenerate")
            if not job_id:
                raise ValueError("No job ID returned from server")

            return str(job_id)

        except Exception as e:
            self.log.error(f"Failed to generate metadata: {e}")
            raise

    async def metadata_scan(
        self,
        paths: list[str] | None = None,
        flags: dict[str, Any] | None = None,
    ) -> str:
        """Start a metadata scan job.

        Args:
            paths: List of paths to scan (None = all paths)
            flags: Dict of scan flags to override defaults (rescan, scanGenerateCovers, etc.)

        Returns:
            Job ID for the scan operation
        """
        # Get scan input object with defaults from config
        if paths is None:
            paths = []
        try:
            defaults = await self.get_configuration_defaults()
            scan_input = ScanMetadataInput(
                paths=paths,
                rescan=getattr(defaults.scan, "rescan", False),
                scanGenerateCovers=getattr(defaults.scan, "scanGenerateCovers", True),
                scanGeneratePreviews=getattr(
                    defaults.scan, "scanGeneratePreviews", True
                ),
                scanGenerateImagePreviews=getattr(
                    defaults.scan, "scanGenerateImagePreviews", True
                ),
                scanGenerateSprites=getattr(defaults.scan, "scanGenerateSprites", True),
                scanGeneratePhashes=getattr(defaults.scan, "scanGeneratePhashes", True),
                scanGenerateThumbnails=getattr(
                    defaults.scan, "scanGenerateThumbnails", True
                ),
                scanGenerateClipPreviews=getattr(
                    defaults.scan, "scanGenerateClipPreviews", True
                ),
            )
        except Exception as e:
            self.log.warning(
                f"Failed to get scan defaults: {e}, using hardcoded defaults"
            )
            scan_input = ScanMetadataInput(
                paths=paths,
                rescan=False,
                scanGenerateCovers=True,
                scanGeneratePreviews=True,
                scanGenerateImagePreviews=True,
                scanGenerateSprites=True,
                scanGeneratePhashes=True,
                scanGenerateThumbnails=True,
                scanGenerateClipPreviews=True,
            )

        # Override with any provided flags
        if flags:
            for key, value in flags.items():
                setattr(scan_input, key, value)

        # Convert to dict for GraphQL
        variables = {"input": scan_input.__dict__}
        try:
            result = await self.execute(fragments.METADATA_SCAN_MUTATION, variables)
            job_id = result.get("metadataScan")
            if not job_id:
                raise ValueError("Failed to start metadata scan - no job ID returned")
            return str(job_id)
        except Exception as e:
            self.log.error(f"Failed to start metadata scan: {e}")
            raise ValueError(f"Failed to start metadata scan: {e}")

    async def get_configuration_defaults(self) -> ConfigDefaultSettingsResult:
        """Get default configuration settings."""
        try:
            # Use new execute() with result_type for automatic deserialization
            config = await self.execute(
                fragments.CONFIG_DEFAULTS_QUERY, result_type=ConfigResult
            )

            # ConfigResult has a defaults field - Pydantic handles all nested objects
            if config and is_set(config.defaults):
                return config.defaults

            # Fallback to defaults
            self.log.warning("No defaults in response, using hardcoded values")
            return ConfigDefaultSettingsResult(
                scan=ScanMetadataOptions(
                    rescan=False,
                    scanGenerateCovers=True,
                    scanGeneratePreviews=True,
                    scanGenerateImagePreviews=True,
                    scanGenerateSprites=True,
                    scanGeneratePhashes=True,
                    scanGenerateThumbnails=True,
                    scanGenerateClipPreviews=True,
                ),
                autoTag=AutoTagMetadataOptions(),
                generate=GenerateMetadataOptions(),
                deleteFile=False,
                deleteGenerated=False,
            )
        except Exception as e:
            self.log.error(f"Failed to get configuration defaults: {e}")
            raise

    async def metadata_clean(
        self,
        input_data: CleanMetadataInput | dict[str, Any],
    ) -> str:
        """Clean metadata and remove orphaned database entries.

        Args:
            input_data: CleanMetadataInput object or dictionary containing:
                - paths: List of paths to clean (optional)
                - dry_run: Whether to perform a dry run (optional, default: False)

        Returns:
            Job ID for the clean operation

        Examples:
            Clean all metadata:
            ```python
            job_id = await client.metadata_clean({"dry_run": False})
            print(f"Clean job started: {job_id}")
            ```

            Dry run to see what would be cleaned:
            ```python
            job_id = await client.metadata_clean({"dry_run": True})
            ```

            Clean specific paths:
            ```python
            from stash_graphql_client.types import CleanMetadataInput

            input_data = CleanMetadataInput(
                paths=["/path/to/clean"],
                dry_run=False
            )
            job_id = await client.metadata_clean(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, CleanMetadataInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be CleanMetadataInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = CleanMetadataInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.METADATA_CLEAN_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("metadataClean", ""))
        except Exception as e:
            self.log.error(f"Failed to clean metadata: {e}")
            raise

    async def metadata_clean_generated(
        self,
        input_data: CleanGeneratedInput | dict[str, Any],
    ) -> str:
        """Clean generated files (sprites, previews, screenshots, etc.).

        Args:
            input_data: CleanGeneratedInput object or dictionary containing:
                - blobFiles: Clean blob files (optional)
                - dryRun: Whether to perform a dry run (optional, default: False)
                - imageThumbnails: Clean image thumbnails (optional)
                - markers: Clean marker files (optional)
                - screenshots: Clean screenshot files (optional)
                - sprites: Clean sprite files (optional)
                - transcodes: Clean transcode files (optional)

        Returns:
            Job ID for the clean operation

        Examples:
            Clean all generated files:
            ```python
            job_id = await client.metadata_clean_generated({
                "blobFiles": True,
                "imageThumbnails": True,
                "markers": True,
                "screenshots": True,
                "sprites": True,
                "transcodes": True,
                "dryRun": False
            })
            ```

            Dry run to see what would be cleaned:
            ```python
            job_id = await client.metadata_clean_generated({"dryRun": True})
            ```

            Clean only specific types:
            ```python
            from stash_graphql_client.types import CleanGeneratedInput

            input_data = CleanGeneratedInput(
                sprites=True,
                screenshots=True,
                dryRun=False
            )
            job_id = await client.metadata_clean_generated(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, CleanGeneratedInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be CleanGeneratedInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = CleanGeneratedInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.METADATA_CLEAN_GENERATED_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("metadataCleanGenerated", ""))
        except Exception as e:
            self.log.error(f"Failed to clean generated files: {e}")
            raise

    async def metadata_auto_tag(
        self,
        input_data: AutoTagMetadataInput | dict[str, Any],
    ) -> str:
        """Start auto-tagging metadata task.

        Args:
            input_data: AutoTagMetadataInput object or dictionary containing:
                - paths: List of paths to tag, None for all files (optional)
                - performers: List of performer IDs to tag with, or ["*"] for all (optional)
                - studios: List of studio IDs to tag with, or ["*"] for all (optional)
                - tags: List of tag IDs to tag with, or ["*"] for all (optional)

        Returns:
            Job ID for the auto-tagging task

        Examples:
            Auto-tag all files with all performers:
            ```python
            job_id = await client.metadata_auto_tag({
                "performers": ["*"]
            })
            print(f"Auto-tag job started: {job_id}")
            ```

            Auto-tag specific paths with specific performers:
            ```python
            from stash_graphql_client.types import AutoTagMetadataInput

            input_data = AutoTagMetadataInput(
                paths=["/path/to/videos"],
                performers=["1", "2", "3"],
                studios=["*"],
                tags=None
            )
            job_id = await client.metadata_auto_tag(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, AutoTagMetadataInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be AutoTagMetadataInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = AutoTagMetadataInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.METADATA_AUTO_TAG_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("metadataAutoTag", ""))
        except Exception as e:
            self.log.error(f"Failed to start auto-tag: {e}")
            raise

    async def metadata_identify(
        self,
        input_data: IdentifyMetadataInput | dict[str, Any],
    ) -> str:
        """Start metadata identification task using scrapers.

        Args:
            input_data: IdentifyMetadataInput object or dictionary containing:
                - sources: List of scraper sources to use (required)
                - options: Identification options (optional)
                - sceneIDs: List of scene IDs to identify (optional)
                - paths: List of scene paths to identify (optional, ignored if sceneIDs set)

        Returns:
            Job ID for the identification task

        Examples:
            Identify scenes using a stash-box endpoint:
            ```python
            job_id = await client.metadata_identify({
                "sources": [
                    {
                        "source": {
                            "stashBoxEndpoint": "https://stashdb.org/graphql"
                        }
                    }
                ],
                "sceneIDs": ["1", "2", "3"]
            })
            print(f"Identify job started: {job_id}")
            ```

            Identify with custom options:
            ```python
            from stash_graphql_client.types import (
                IdentifyMetadataInput,
                IdentifySourceInput,
                ScraperSourceInput,
                IdentifyMetadataOptionsInput
            )

            input_data = IdentifyMetadataInput(
                sources=[
                    IdentifySourceInput(
                        source=ScraperSourceInput(
                            stashBoxEndpoint="https://stashdb.org/graphql"
                        ),
                        options=IdentifyMetadataOptionsInput(
                            setCoverImage=True,
                            includeMalePerformers=False
                        )
                    )
                ],
                paths=["/path/to/scenes"]
            )
            job_id = await client.metadata_identify(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, IdentifyMetadataInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be IdentifyMetadataInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = IdentifyMetadataInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.METADATA_IDENTIFY_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("metadataIdentify", ""))
        except Exception as e:
            self.log.error(f"Failed to start identify: {e}")
            raise

    async def metadata_import(self) -> str:
        """Start full metadata import from metadata directory.

        This operation completely wipes the database and imports from the
        metadata directory. Use with caution.

        Returns:
            Job ID for the import task

        Examples:
            Start full metadata import:
            ```python
            job_id = await client.metadata_import()
            print(f"Import job started: {job_id}")
            ```
        """
        try:
            result = await self.execute(fragments.METADATA_IMPORT_MUTATION, {})
            return str(result.get("metadataImport", ""))
        except Exception as e:
            self.log.error(f"Failed to start metadata import: {e}")
            raise

    async def metadata_export(self) -> str:
        """Start full metadata export to metadata directory.

        Exports the entire database to the configured metadata directory.

        Returns:
            Job ID for the export task

        Examples:
            Start full metadata export:
            ```python
            job_id = await client.metadata_export()
            print(f"Export job started: {job_id}")
            ```
        """
        try:
            result = await self.execute(fragments.METADATA_EXPORT_MUTATION, {})
            return str(result.get("metadataExport", ""))
        except Exception as e:
            self.log.error(f"Failed to start metadata export: {e}")
            raise

    async def export_objects(
        self,
        input_data: ExportObjectsInput | dict[str, Any],
    ) -> str:
        """Export objects to a downloadable file.

        Args:
            input_data: ExportObjectsInput object or dictionary containing:
                - ids: List of object IDs to export (optional)
                - all: Export all objects (optional, default: False)
                - type: Object type to export (required)
                - format: Export format (optional)

        Returns:
            Download token for the exported file

        Examples:
            Export all scenes:
            ```python
            token = await client.export_objects({
                "all": True,
                "type": "SCENE"
            })
            download_url = f"{client.url}/downloads/{token}"
            ```

            Export specific performers:
            ```python
            from stash_graphql_client.types import ExportObjectsInput

            input_data = ExportObjectsInput(
                ids=["1", "2", "3"],
                type="PERFORMER"
            )
            token = await client.export_objects(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, ExportObjectsInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be ExportObjectsInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = ExportObjectsInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.EXPORT_OBJECTS_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("exportObjects", ""))
        except Exception as e:
            self.log.error(f"Failed to export objects: {e}")
            raise

    async def import_objects(
        self,
        input_data: ImportObjectsInput | dict[str, Any],
    ) -> str:
        """Import objects from a file.

        Args:
            input_data: ImportObjectsInput object or dictionary containing:
                - file: File to import from (required)
                - duplicateBehaviour: How to handle duplicates (optional)
                - missingRefBehaviour: How to handle missing references (optional)

        Returns:
            Import job ID

        Examples:
            Import from file:
            ```python
            job_id = await client.import_objects({
                "file": "/path/to/export.json"
            })
            print(f"Import job started: {job_id}")
            ```

            Import with duplicate handling:
            ```python
            from stash_graphql_client.types import ImportObjectsInput

            input_data = ImportObjectsInput(
                file="/path/to/export.json",
                duplicateBehaviour="IGNORE"
            )
            job_id = await client.import_objects(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, ImportObjectsInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be ImportObjectsInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = ImportObjectsInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.IMPORT_OBJECTS_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("importObjects", ""))
        except Exception as e:
            self.log.error(f"Failed to import objects: {e}")
            raise

    async def backup_database(
        self,
        input_data: BackupDatabaseInput | dict[str, Any],
    ) -> str:
        """Create a database backup.

        Args:
            input_data: BackupDatabaseInput object or dictionary containing:
                - download: Whether to download the backup (optional, default: True)

        Returns:
            Backup file path or download token

        Examples:
            Create and download backup:
            ```python
            token = await client.backup_database({"download": True})
            download_url = f"{client.url}/downloads/{token}"
            ```

            Create backup without downloading:
            ```python
            from stash_graphql_client.types import BackupDatabaseInput

            input_data = BackupDatabaseInput(download=False)
            path = await client.backup_database(input_data)
            print(f"Backup created at: {path}")
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, BackupDatabaseInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be BackupDatabaseInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = BackupDatabaseInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.BACKUP_DATABASE_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("backupDatabase", ""))
        except Exception as e:
            self.log.error(f"Failed to backup database: {e}")
            raise

    async def anonymise_database(
        self,
        input_data: AnonymiseDatabaseInput | dict[str, Any],
    ) -> str:
        """Anonymise the database by removing identifying information.

        Args:
            input_data: AnonymiseDatabaseInput object or dictionary containing:
                - download: Whether to download the anonymised backup (optional, default: True)

        Returns:
            Anonymised backup file path or download token

        Examples:
            Anonymise and download:
            ```python
            token = await client.anonymise_database({"download": True})
            download_url = f"{client.url}/downloads/{token}"
            ```

            Anonymise without downloading:
            ```python
            from stash_graphql_client.types import AnonymiseDatabaseInput

            input_data = AnonymiseDatabaseInput(download=False)
            path = await client.anonymise_database(input_data)
            print(f"Anonymised backup created at: {path}")
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, AnonymiseDatabaseInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be AnonymiseDatabaseInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = AnonymiseDatabaseInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.ANONYMISE_DATABASE_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("anonymiseDatabase", ""))
        except Exception as e:
            self.log.error(f"Failed to anonymise database: {e}")
            raise

    async def migrate(
        self,
        input_data: MigrateInput | dict[str, Any],
    ) -> str:
        """Migrate database to the latest schema version.

        Args:
            input_data: MigrateInput object or dictionary containing:
                - backupPath: Path to create backup before migration (required)

        Returns:
            Migration job ID

        Examples:
            Migrate database with backup:
            ```python
            job_id = await client.migrate({"backupPath": "/path/to/backup.db"})
            print(f"Migration job started: {job_id}")
            ```

            Using the input type:
            ```python
            from stash_graphql_client.types import MigrateInput

            input_data = MigrateInput(backupPath="/path/to/backup.db")
            job_id = await client.migrate(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, MigrateInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be MigrateInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = MigrateInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.MIGRATE_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("migrate", ""))
        except Exception as e:
            self.log.error(f"Failed to migrate database: {e}")
            raise

    async def migrate_hash_naming(self) -> str:
        """Migrate hash naming scheme to the latest version.

        Returns:
            Migration job ID

        Examples:
            Migrate hash naming:
            ```python
            job_id = await client.migrate_hash_naming()
            print(f"Hash naming migration started: {job_id}")
            ```
        """
        try:
            result = await self.execute(
                fragments.MIGRATE_HASH_NAMING_MUTATION,
                {},
            )
            return str(result.get("migrateHashNaming", ""))
        except Exception as e:
            self.log.error(f"Failed to migrate hash naming: {e}")
            raise

    async def migrate_scene_screenshots(
        self,
        input_data: MigrateSceneScreenshotsInput | dict[str, Any],
    ) -> str:
        """Migrate scene screenshots to the latest storage format.

        Args:
            input_data: MigrateSceneScreenshotsInput object or dictionary containing:
                - deleteFiles: Delete old screenshot files (optional)
                - overwriteExisting: Overwrite existing screenshots (optional)

        Returns:
            Migration job ID

        Examples:
            Migrate screenshots and delete old files:
            ```python
            job_id = await client.migrate_scene_screenshots({
                "deleteFiles": True,
                "overwriteExisting": False
            })
            print(f"Screenshot migration started: {job_id}")
            ```

            Using the input type:
            ```python
            from stash_graphql_client.types import MigrateSceneScreenshotsInput

            input_data = MigrateSceneScreenshotsInput(
                deleteFiles=True,
                overwriteExisting=True
            )
            job_id = await client.migrate_scene_screenshots(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, MigrateSceneScreenshotsInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be MigrateSceneScreenshotsInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = MigrateSceneScreenshotsInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.MIGRATE_SCENE_SCREENSHOTS_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("migrateSceneScreenshots", ""))
        except Exception as e:
            self.log.error(f"Failed to migrate scene screenshots: {e}")
            raise

    async def migrate_blobs(
        self,
        input_data: MigrateBlobsInput | dict[str, Any],
    ) -> str:
        """Migrate blobs to the latest storage format.

        Args:
            input_data: MigrateBlobsInput object or dictionary containing:
                - deleteOld: Delete old blob files after migration (optional)

        Returns:
            Migration job ID

        Examples:
            Migrate blobs and keep old files:
            ```python
            job_id = await client.migrate_blobs({"deleteOld": False})
            print(f"Blob migration started: {job_id}")
            ```

            Using the input type:
            ```python
            from stash_graphql_client.types import MigrateBlobsInput

            input_data = MigrateBlobsInput(deleteOld=True)
            job_id = await client.migrate_blobs(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, MigrateBlobsInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be MigrateBlobsInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = MigrateBlobsInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.MIGRATE_BLOBS_MUTATION,
                {"input": input_dict},
            )
            return str(result.get("migrateBlobs", ""))
        except Exception as e:
            self.log.error(f"Failed to migrate blobs: {e}")
            raise

    async def optimise_database(self) -> str:
        """Optimize the database.

        Returns:
            Job ID for the optimization task
        """
        try:
            result = await self.execute(fragments.OPTIMISE_DATABASE_MUTATION, {})
            return str(result.get("optimiseDatabase", ""))
        except Exception as e:
            self.log.error(f"Failed to optimize database: {e}")
            raise

    async def setup(
        self,
        input_data: SetupInput | dict[str, Any],
    ) -> bool:
        """Run initial Stash setup.

        Args:
            input_data: SetupInput object or dictionary

        Returns:
            True if successful
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, SetupInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be SetupInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = SetupInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.SETUP_MUTATION,
                {"input": input_dict},
            )
            return result.get("setup") is True
        except Exception as e:
            self.log.error(f"Failed to run setup: {e}")
            raise

    async def download_ffmpeg(self) -> str:
        """Download FFmpeg binary.

        Returns:
            Job ID for the download task
        """
        try:
            result = await self.execute(fragments.DOWNLOAD_FFMPEG_MUTATION, {})
            return str(result.get("downloadFFMPEG", ""))
        except Exception as e:
            self.log.error(f"Failed to download FFmpeg: {e}")
            raise



"""Mixin for operations that are intentionally not implemented.

This mixin contains stub methods for GraphQL operations that are either:
1. Security risks (direct SQL access)
2. Not yet prioritized for implementation (DLNA, StashBox validation)

Due to Python's Method Resolution Order (MRO), this mixin is placed LAST
in the StashClient inheritance chain, so any real implementations in other
mixins will automatically override these stubs.

IMPORTANT: If you implement any of these methods in another mixin, you should
remove the corresponding stub from this file to avoid confusion.
"""

from typing import Any

from ...types import (
    DLNAStatus,
    SQLExecResult,
    SQLQueryResult,
    StashBoxValidationResult,
)


class NotImplementedClientMixin:
    """Mixin containing stubs for unimplemented GraphQL operations.

    This class serves as:
    - Documentation of which operations exist in the schema but aren't implemented
    - Type hints for IDE autocomplete
    - Clear error messages when calling unimplemented operations

    All methods here raise NotImplementedError when called.
    """

    # =========================================================================
    # SQL Operations - INTENTIONALLY NOT IMPLEMENTED (Security Risk)
    # =========================================================================
    # Direct SQL access poses significant security risks:
    # - SQL injection vulnerabilities
    # - Potential for data corruption
    # - Bypasses application-level validation and business logic
    # - Could expose sensitive data
    #
    # Users requiring SQL access should use Stash's UI or direct database tools.

    async def querySQL(self, sql: str, args: list[Any] | None = None) -> SQLQueryResult:
        """Execute SQL query.

        SECURITY: This operation is intentionally not implemented to prevent
        SQL injection vulnerabilities and unauthorized database access.

        Args:
            sql: SQL query string
            args: Optional query arguments

        Raises:
            NotImplementedError: Always raised (security measure)
        """
        raise NotImplementedError(
            "SQL queries are not implemented for security reasons. "
            "Use Stash's UI or direct database tools for SQL access."
        )

    async def execSQL(self, sql: str, args: list[Any] | None = None) -> SQLExecResult:
        """Execute SQL statement.

        SECURITY: This operation is intentionally not implemented to prevent
        SQL injection vulnerabilities and unauthorized database modifications.

        Args:
            sql: SQL statement string
            args: Optional statement arguments

        Raises:
            NotImplementedError: Always raised (security measure)
        """
        raise NotImplementedError(
            "SQL execution is not implemented for security reasons. "
            "Use Stash's UI or direct database tools for SQL access."
        )

    # =========================================================================
    # StashBox Operations - Not Yet Implemented
    # =========================================================================

    async def validateStashBoxCredentials(
        self, input_data: dict[str, Any]
    ) -> StashBoxValidationResult:
        """Validate StashBox instance credentials.

        Tests whether the provided StashBox credentials are valid.

        Args:
            input_data: Dictionary containing StashBox endpoint and API key

        Returns:
            StashBoxValidationResult with validation status

        Raises:
            NotImplementedError: Not yet implemented
        """
        raise NotImplementedError(
            "StashBox credential validation is not yet implemented. "
            "Please validate credentials through Stash's UI."
        )

    # =========================================================================
    # DLNA Operations - Not Yet Implemented
    # =========================================================================

    async def dlnaStatus(self) -> DLNAStatus:
        """Get DLNA server status.

        Returns information about the DLNA server state, including
        whether it's running and what IPs are whitelisted.

        Returns:
            DLNAStatus with server information

        Raises:
            NotImplementedError: Not yet implemented
        """
        raise NotImplementedError(
            "DLNA status query is not yet implemented. "
            "Please check DLNA status through Stash's UI."
        )



"""Metadata types from schema/types/metadata.graphql."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field

from .base import FromGraphQLMixin, StashInput
from .enums import (
    IdentifyFieldStrategy,
    ImportDuplicateEnum,
    ImportMissingRefEnum,
    PreviewPreset,
    SystemStatusEnum,
)
from .unset import UNSET, UnsetType


if TYPE_CHECKING:
    from .scraped_types import ScraperSource, ScraperSourceInput


class GeneratePreviewOptionsInput(StashInput):
    """Input for preview generation options from schema/types/metadata.graphql."""

    previewSegments: int | None | UnsetType = (
        UNSET  # Int (Number of segments in a preview file)
    )
    previewSegmentDuration: float | None | UnsetType = (
        UNSET  # Float (Preview segment duration, in seconds)
    )
    previewExcludeStart: str | None | UnsetType = (
        UNSET  # String (Duration of start of video to exclude when generating previews)
    )
    previewExcludeEnd: str | None | UnsetType = (
        UNSET  # String (Duration of end of video to exclude when generating previews)
    )
    previewPreset: PreviewPreset | None | UnsetType = (
        UNSET  # PreviewPreset (Preset when generating preview)
    )


class GenerateMetadataInput(StashInput):
    """Input for metadata generation from schema/types/metadata.graphql."""

    covers: bool | UnsetType = UNSET  # Boolean
    sprites: bool | UnsetType = UNSET  # Boolean
    previews: bool | UnsetType = UNSET  # Boolean
    imagePreviews: bool | UnsetType = UNSET  # Boolean
    previewOptions: GeneratePreviewOptionsInput | None | UnsetType = (
        UNSET  # GeneratePreviewOptionsInput
    )
    markers: bool | UnsetType = UNSET  # Boolean
    markerImagePreviews: bool | UnsetType = UNSET  # Boolean
    markerScreenshots: bool | UnsetType = UNSET  # Boolean
    transcodes: bool | UnsetType = UNSET  # Boolean
    forceTranscodes: bool | UnsetType = UNSET  # Boolean
    phashes: bool | UnsetType = UNSET  # Boolean
    interactiveHeatmapsSpeeds: bool | UnsetType = UNSET  # Boolean
    imageThumbnails: bool | UnsetType = UNSET  # Boolean
    clipPreviews: bool | UnsetType = UNSET  # Boolean
    sceneIDs: list[str] | None | UnsetType = UNSET  # [ID!] (scene ids to generate for)
    markerIDs: list[str] | None | UnsetType = (
        UNSET  # [ID!] (marker ids to generate for)
    )
    overwrite: bool | UnsetType = UNSET  # Boolean (overwrite existing media)


class GeneratePreviewOptions(BaseModel):
    """Preview generation options from schema/types/metadata.graphql."""

    previewSegments: int | None | UnsetType = (
        UNSET  # Int (Number of segments in a preview file)
    )
    previewSegmentDuration: float | None | UnsetType = (
        UNSET  # Float (Preview segment duration, in seconds)
    )
    previewExcludeStart: str | None | UnsetType = (
        UNSET  # String (Duration of start of video to exclude when generating previews)
    )
    previewExcludeEnd: str | None | UnsetType = (
        UNSET  # String (Duration of end of video to exclude when generating previews)
    )
    previewPreset: PreviewPreset | None | UnsetType = (
        UNSET  # PreviewPreset (Preset when generating preview)
    )


class GenerateMetadataOptions(BaseModel):
    """Metadata generation options from schema/types/metadata.graphql."""

    covers: bool | None | UnsetType = UNSET  # Boolean
    sprites: bool | None | UnsetType = UNSET  # Boolean
    previews: bool | None | UnsetType = UNSET  # Boolean
    imagePreviews: bool | None | UnsetType = UNSET  # Boolean
    previewOptions: GeneratePreviewOptions | None | UnsetType = (
        UNSET  # GeneratePreviewOptions
    )
    markers: bool | None | UnsetType = UNSET  # Boolean
    markerImagePreviews: bool | None | UnsetType = UNSET  # Boolean
    markerScreenshots: bool | None | UnsetType = UNSET  # Boolean
    transcodes: bool | None | UnsetType = UNSET  # Boolean
    phashes: bool | None | UnsetType = UNSET  # Boolean
    interactiveHeatmapsSpeeds: bool | None | UnsetType = UNSET  # Boolean
    imageThumbnails: bool | None | UnsetType = UNSET  # Boolean
    clipPreviews: bool | None | UnsetType = UNSET  # Boolean


class ScanMetaDataFilterInput(StashInput):
    """Filter options for meta data scanning from schema/types/metadata.graphql."""

    minModTime: datetime | None | UnsetType = (
        UNSET  # Timestamp (If set, files with a modification time before this time point are ignored by the scan)
    )


class ScanMetadataInput(StashInput):
    """Input for metadata scanning from schema/types/metadata.graphql."""

    paths: list[str] | UnsetType = UNSET  # [String!]
    rescan: bool | None | UnsetType = (
        UNSET  # Boolean (Forces a rescan on files even if modification time is unchanged)
    )
    scanGenerateCovers: bool | None | UnsetType = (
        UNSET  # Boolean (Generate covers during scan)
    )
    scanGeneratePreviews: bool | None | UnsetType = (
        UNSET  # Boolean (Generate previews during scan)
    )
    scanGenerateImagePreviews: bool | None | UnsetType = (
        UNSET  # Boolean (Generate image previews during scan)
    )
    scanGenerateSprites: bool | None | UnsetType = (
        UNSET  # Boolean (Generate sprites during scan)
    )
    scanGeneratePhashes: bool | None | UnsetType = (
        UNSET  # Boolean (Generate phashes during scan)
    )
    scanGenerateThumbnails: bool | None | UnsetType = (
        UNSET  # Boolean (Generate image thumbnails during scan)
    )
    scanGenerateClipPreviews: bool | None | UnsetType = (
        UNSET  # Boolean (Generate image clip previews during scan)
    )
    filter: ScanMetaDataFilterInput | None | UnsetType = (
        UNSET  # ScanMetaDataFilterInput (Filter options for the scan)
    )


class ScanMetadataOptions(BaseModel):
    """Metadata scan options from schema/types/metadata.graphql."""

    rescan: bool | UnsetType = (
        UNSET  # Boolean! (Forces a rescan on files even if modification time is unchanged)
    )
    scanGenerateCovers: bool | UnsetType = (
        UNSET  # Boolean! (Generate covers during scan)
    )
    scanGeneratePreviews: bool | UnsetType = (
        UNSET  # Boolean! (Generate previews during scan)
    )
    scanGenerateImagePreviews: bool | UnsetType = (
        UNSET  # Boolean! (Generate image previews during scan)
    )
    scanGenerateSprites: bool | UnsetType = (
        UNSET  # Boolean! (Generate sprites during scan)
    )
    scanGeneratePhashes: bool | UnsetType = (
        UNSET  # Boolean! (Generate phashes during scan)
    )
    scanGenerateThumbnails: bool | UnsetType = (
        UNSET  # Boolean! (Generate image thumbnails during scan)
    )
    scanGenerateClipPreviews: bool | UnsetType = (
        UNSET  # Boolean! (Generate image clip previews during scan)
    )


class CleanMetadataInput(StashInput):
    """Input for metadata cleaning from schema/types/metadata.graphql."""

    paths: list[str] | UnsetType = UNSET  # [String!]
    dry_run: bool | UnsetType = Field(
        default=UNSET, alias="dryRun"
    )  # Boolean! (Do a dry run. Don't delete any files)


class CleanGeneratedInput(StashInput):
    """Input for cleaning generated files from schema/types/metadata.graphql."""

    blob_files: bool | None | UnsetType = Field(
        default=UNSET, alias="blobFiles"
    )  # Boolean (Clean blob files without blob entries)
    sprites: bool | None | UnsetType = (
        UNSET  # Boolean (Clean sprite and vtt files without scene entries)
    )
    screenshots: bool | None | UnsetType = (
        UNSET  # Boolean (Clean preview files without scene entries)
    )
    transcodes: bool | None | UnsetType = (
        UNSET  # Boolean (Clean scene transcodes without scene entries)
    )
    markers: bool | None | UnsetType = (
        UNSET  # Boolean (Clean marker files without marker entries)
    )
    image_thumbnails: bool | None | UnsetType = Field(
        default=UNSET, alias="imageThumbnails"
    )  # Boolean (Clean image thumbnails/clips without image entries)
    dry_run: bool | None | UnsetType = Field(
        default=UNSET, alias="dryRun"
    )  # Boolean (Do a dry run. Don't delete any files)


class AutoTagMetadataInput(StashInput):
    """Input for auto-tagging metadata from schema/types/metadata.graphql."""

    paths: list[str] | None | UnsetType = (
        UNSET  # [String!] (Paths to tag, null for all files)
    )
    performers: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of performers to tag files with, or "*" for all)
    )
    studios: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of studios to tag files with, or "*" for all)
    )
    tags: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of tags to tag files with, or "*" for all)
    )


class AutoTagMetadataOptions(BaseModel):
    """Auto-tag metadata options from schema/types/metadata.graphql."""

    performers: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of performers to tag files with, or "*" for all)
    )
    studios: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of studios to tag files with, or "*" for all)
    )
    tags: list[str] | None | UnsetType = (
        UNSET  # [String!] (IDs of tags to tag files with, or "*" for all)
    )


class IdentifyFieldOptionsInput(StashInput):
    """Input for identify field options from schema/types/metadata.graphql."""

    field: str | UnsetType = UNSET  # String!
    strategy: IdentifyFieldStrategy | UnsetType = UNSET  # IdentifyFieldStrategy!
    createMissing: bool | None | UnsetType = (
        UNSET  # Boolean (creates missing objects if needed - only applicable for performers, tags and studios)
    )


class IdentifyMetadataOptionsInput(StashInput):
    """Input for identify metadata options from schema/types/metadata.graphql."""

    fieldOptions: list[IdentifyFieldOptionsInput] | None | UnsetType = (
        UNSET  # [IdentifyFieldOptionsInput!] (any fields missing from here are defaulted to MERGE and createMissing false)
    )
    setCoverImage: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    setOrganized: bool | None | UnsetType = UNSET  # Boolean
    includeMalePerformers: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipMultipleMatches: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipMultipleMatchTag: str | None | UnsetType = (
        UNSET  # String (tag to tag skipped multiple matches with)
    )
    skipSingleNamePerformers: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipSingleNamePerformerTag: str | None | UnsetType = (
        UNSET  # String (tag to tag skipped single name performers with)
    )


class IdentifyFieldOptions(BaseModel):
    """Identify field options from schema/types/metadata.graphql."""

    field: str | UnsetType = UNSET  # String!
    strategy: IdentifyFieldStrategy | UnsetType = UNSET  # IdentifyFieldStrategy!
    createMissing: bool | UnsetType = (
        UNSET  # Boolean (creates missing objects if needed - only applicable for performers, tags and studios)
    )


class IdentifyMetadataOptions(BaseModel):
    """Identify metadata options from schema/types/metadata.graphql."""

    fieldOptions: list[IdentifyFieldOptions] | None | UnsetType = (
        UNSET  # [IdentifyFieldOptions!] (any fields missing from here are defaulted to MERGE and createMissing false)
    )
    setCoverImage: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    setOrganized: bool | None | UnsetType = UNSET  # Boolean
    includeMalePerformers: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipMultipleMatches: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipMultipleMatchTag: str | None | UnsetType = (
        UNSET  # String (tag to tag skipped multiple matches with)
    )
    skipSingleNamePerformers: bool | None | UnsetType = (
        UNSET  # Boolean (defaults to true if not provided)
    )
    skipSingleNamePerformerTag: str | None | UnsetType = (
        UNSET  # String (tag to tag skipped single name performers with)
    )


class ExportObjectTypeInput(StashInput):
    """Input for export object type from schema/types/metadata.graphql."""

    ids: list[str] | None | UnsetType = UNSET  # [String!]
    all: bool | None | UnsetType = UNSET  # Boolean


class ExportObjectsInput(StashInput):
    """Input for exporting objects from schema/types/metadata.graphql."""

    scenes: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    images: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    studios: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    performers: ExportObjectTypeInput | None | UnsetType = (
        UNSET  # ExportObjectTypeInput
    )
    tags: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    groups: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    galleries: ExportObjectTypeInput | None | UnsetType = UNSET  # ExportObjectTypeInput
    include_dependencies: bool | None | UnsetType = Field(
        default=UNSET, alias="includeDependencies"
    )  # Boolean


class ImportObjectsInput(StashInput):
    """Input for importing objects from schema/types/metadata.graphql."""

    file: Any | UnsetType = UNSET  # Upload!
    duplicate_behaviour: ImportDuplicateEnum | UnsetType = Field(
        default=UNSET, alias="duplicateBehaviour"
    )  # ImportDuplicateEnum!
    missing_ref_behaviour: ImportMissingRefEnum | UnsetType = Field(
        default=UNSET, alias="missingRefBehaviour"
    )  # ImportMissingRefEnum!


class BackupDatabaseInput(StashInput):
    """Input for database backup from schema/types/metadata.graphql."""

    download: bool | None | UnsetType = UNSET  # Boolean


class AnonymiseDatabaseInput(StashInput):
    """Input for database anonymisation from schema/types/metadata.graphql."""

    download: bool | None | UnsetType = UNSET  # Boolean


class SystemStatus(FromGraphQLMixin, BaseModel):
    """System status type from schema/types/metadata.graphql."""

    databaseSchema: int | None | UnsetType = UNSET  # Int
    databasePath: str | None | UnsetType = UNSET  # String
    configPath: str | None | UnsetType = UNSET  # String
    appSchema: int | UnsetType = UNSET  # Int!
    status: SystemStatusEnum | UnsetType = UNSET  # SystemStatusEnum!
    os: str | UnsetType = UNSET  # String!
    workingDir: str | UnsetType = UNSET  # String!
    homeDir: str | UnsetType = UNSET  # String!
    ffmpegPath: str | None | UnsetType = UNSET  # String
    ffprobePath: str | None | UnsetType = UNSET  # String


class MigrateInput(StashInput):
    """Input for migration from schema/types/metadata.graphql."""

    backup_path: str | UnsetType = Field(default=UNSET, alias="backupPath")  # String!


class CustomFieldsInput(StashInput):
    """Input for custom fields from schema/types/metadata.graphql."""

    full: dict[str, Any] | None | UnsetType = (
        UNSET  # Map (If populated, the entire custom fields map will be replaced with this value)
    )
    partial: dict[str, Any] | None | UnsetType = (
        UNSET  # Map (If populated, only the keys in this map will be updated)
    )


class IdentifySourceInput(StashInput):
    """Input for identify source from schema/types/metadata.graphql."""

    source: ScraperSourceInput | UnsetType = UNSET  # ScraperSourceInput!
    options: IdentifyMetadataOptionsInput | None | UnsetType = (
        UNSET  # IdentifyMetadataOptionsInput (Options defined for a source override the defaults)
    )


class IdentifySource(BaseModel):
    """Identify source type from schema/types/metadata.graphql."""

    source: ScraperSource | UnsetType = UNSET  # ScraperSource!
    options: IdentifyMetadataOptions | None | UnsetType = (
        UNSET  # IdentifyMetadataOptions (Options defined for a source override the defaults)
    )


class IdentifyMetadataInput(StashInput):
    """Input for identify metadata from schema/types/metadata.graphql."""

    sources: list[IdentifySourceInput] | UnsetType = (
        UNSET  # [IdentifySourceInput!]! (An ordered list of sources to identify items with. Only the first source that finds a match is used.)
    )
    options: IdentifyMetadataOptionsInput | None | UnsetType = (
        UNSET  # IdentifyMetadataOptionsInput (Options defined here override the configured defaults)
    )
    sceneIDs: list[str] | None | UnsetType = UNSET  # [ID!] (scene ids to identify)
    paths: list[str] | None | UnsetType = (
        UNSET  # [String!] (paths of scenes to identify - ignored if scene ids are set)
    )


class IdentifyMetadataTaskOptions(BaseModel):
    """Identify metadata task options from schema/types/metadata.graphql."""

    sources: list[IdentifySource] | UnsetType = (
        UNSET  # [IdentifySource!]! (An ordered list of sources to identify items with. Only the first source that finds a match is used.)
    )
    options: IdentifyMetadataOptions | None | UnsetType = (
        UNSET  # IdentifyMetadataOptions (Options defined here override the configured defaults)
    )


class MigrateSceneScreenshotsInput(StashInput):
    """Input for migrating scene screenshots from schema/types/migration.graphql."""

    delete_files: bool | None | UnsetType = Field(
        UNSET, alias="deleteFiles"
    )  # Boolean (if true, delete screenshot files after migrating)
    overwrite_existing: bool | None | UnsetType = Field(
        UNSET, alias="overwriteExisting"
    )  # Boolean (if true, overwrite existing covers with the covers from the screenshots directory)


class MigrateBlobsInput(StashInput):
    """Input for migrating blobs from schema/types/migration.graphql."""

    delete_old: bool | None | UnsetType = Field(
        UNSET, alias="deleteOld"
    )  # Boolean (if true, delete blob data from old storage system)



"""Stash GraphQL Client - Async Python client for Stash media server.

This library provides a fully async GraphQL client for interacting with
Stash (https://stashapp.cc), a self-hosted media organizer.

Features:
- Async-first design with gql + HTTPXAsyncTransport + WebsocketsTransport
- Pydantic types for all Stash GraphQL schema objects
- Full CRUD operations for all entity types
- Job management and metadata scanning
- GraphQL subscription support for real-time updates

Example:
    >>> from stash_graphql_client import StashClient, StashContext
    >>>
    >>> async with StashContext(conn={"Host": "localhost", "Port": 9999}) as client:
    ...     studios = await client.find_studios()
    ...     print(f"Found {studios.count} studios")
"""

from importlib.metadata import PackageNotFoundError, version

from stash_graphql_client.client import StashClient
from stash_graphql_client.context import StashContext
from stash_graphql_client.logging import (
    client_logger,
    configure_logging,
    processing_logger,
    stash_logger,
)
from stash_graphql_client.store import CacheEntry, StashEntityStore

# Re-export commonly used types
from stash_graphql_client.types import (
    # Base types
    BulkUpdateIds,
    BulkUpdateStrings,
    # Core types
    Gallery,
    GalleryCreateInput,
    GalleryUpdateInput,
    # Metadata types
    GenerateMetadataInput,
    GenerateMetadataOptions,
    Group,
    GroupCreateInput,
    GroupUpdateInput,
    Image,
    Job,
    JobStatus,
    Performer,
    PerformerCreateInput,
    PerformerUpdateInput,
    ScanMetadataInput,
    ScanMetadataOptions,
    Scene,
    SceneCreateInput,
    SceneUpdateInput,
    StashObject,
    Studio,
    StudioCreateInput,
    StudioUpdateInput,
    Tag,
    TagCreateInput,
    TagUpdateInput,
)


# Version is automatically synced from pyproject.toml via poetry-dynamic-versioning
try:
    __version__ = version("stash-graphql-client")
except PackageNotFoundError:
    # Fallback for development installs without package metadata
    __version__ = "0.0.0.dev0"

__all__ = [
    # Version
    "__version__",
    # Client
    "StashClient",
    "StashContext",
    # Store
    "StashEntityStore",
    "CacheEntry",
    # Logging
    "stash_logger",
    "client_logger",
    "processing_logger",
    "configure_logging",
    # Core types
    "Gallery",
    "GalleryCreateInput",
    "GalleryUpdateInput",
    "Group",
    "GroupCreateInput",
    "GroupUpdateInput",
    "Image",
    "Job",
    "JobStatus",
    "Performer",
    "PerformerCreateInput",
    "PerformerUpdateInput",
    "Scene",
    "SceneCreateInput",
    "SceneUpdateInput",
    "Studio",
    "StudioCreateInput",
    "StudioUpdateInput",
    "Tag",
    "TagCreateInput",
    "TagUpdateInput",
    # Base types
    "BulkUpdateIds",
    "BulkUpdateStrings",
    "StashObject",
    # Metadata types
    "GenerateMetadataInput",
    "GenerateMetadataOptions",
    "ScanMetadataInput",
    "ScanMetadataOptions",
]



"""Performer-related client functionality."""

from typing import Any

from ... import fragments
from ...errors import StashGraphQLError
from ...types import (
    BulkPerformerUpdateInput,
    FindPerformersResultType,
    OnMultipleMatch,
    Performer,
    PerformerDestroyInput,
    PerformerMergeInput,
)
from ...types.unset import UnsetType, is_set
from ..protocols import StashClientProtocol


class PerformerClientMixin(StashClientProtocol):
    """Mixin for performer-related client methods."""

    async def find_performer(
        self,
        performer: int | str | dict,
    ) -> Performer | None:
        """Find a performer by ID, name, or filter.

        Args:
            performer: Can be:
                - ID (int/str): Find by ID
                - Name (str): Find by name
                - Dict: Find by filter criteria

        Returns:
            Performer object if found, None otherwise

        Examples:
            Find by ID:
            ```python
            performer = await client.find_performer("123")
            if performer:
                print(f"Found performer: {performer.name}")
            ```

            Find by name:
            ```python
            performer = await client.find_performer("Performer Name")
            if performer:
                print(f"Found performer with ID: {performer.id}")
            ```

            Find by filter:
            ```python
            performer = await client.find_performer({
                "name": "Performer Name",
                "disambiguation": "2000s"
            })
            ```

            Access performer relationships:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Get scene titles
                scene_titles = [s.title for s in performer.scenes]
                # Get studio name
                studio_name = performer.studio.name if performer.studio else None
                # Get tag names
                tags = [t.name for t in performer.tags]
            ```
        """
        try:
            # Parse input to handle different types
            parsed_input = self._parse_obj_for_ID(performer)
            result = None  # Initialize to avoid unbound variable warning

            if isinstance(parsed_input, dict):
                # If it's a name filter, try name then alias
                name = parsed_input.get("name")
                if name:
                    # Try by name first
                    result = await self.find_performers(
                        performer_filter={"name": {"value": name, "modifier": "EQUALS"}}
                    )
                    performers = result.performers
                    count = result.count
                    if (
                        not isinstance(count, UnsetType)
                        and count > 0
                        and not isinstance(performers, UnsetType)
                    ):
                        return performers[0]

                    # Try by alias
                    result = await self.find_performers(
                        performer_filter={
                            "aliases": {"value": name, "modifier": "INCLUDES"}
                        }
                    )
                    performers = result.performers
                    count = result.count
                    if (
                        not isinstance(count, UnsetType)
                        and count > 0
                        and not isinstance(performers, UnsetType)
                    ):
                        return performers[0]

                    return None
            else:
                # If it's an ID, use direct lookup
                raw_result = await self.execute(
                    fragments.FIND_PERFORMER_QUERY,
                    {"id": str(parsed_input)},
                )
                if raw_result and raw_result.get("findPerformer"):
                    return self._decode_result(Performer, raw_result["findPerformer"])
                return None
            return None
        except Exception as e:
            self.log.error(f"Failed to find performer {performer}: {e}")
            return None

    async def find_performers(
        self,
        filter_: dict[str, Any] | None = None,
        performer_filter: dict[str, Any] | None = None,
        q: str | None = None,
    ) -> FindPerformersResultType:
        """Find performers matching the given filters.

        Args:
            filter_: Optional general filter parameters:
                - q: str (search query)
                - direction: SortDirectionEnum (ASC/DESC)
                - page: int
                - per_page: int
                - sort: str (field to sort by)
            q: Optional search query (alternative to filter_["q"])
            performer_filter: Optional performer-specific filter:
                - birth_year: IntCriterionInput
                - age: IntCriterionInput
                - ethnicity: StringCriterionInput
                - country: StringCriterionInput
                - eye_color: StringCriterionInput
                - height: StringCriterionInput
                - measurements: StringCriterionInput
                - fake_tits: StringCriterionInput
                - career_length: StringCriterionInput
                - tattoos: StringCriterionInput
                - piercings: StringCriterionInput
                - favorite: bool
                - rating100: IntCriterionInput
                - gender: GenderEnum
                - is_missing: str (what data is missing)
                - name: StringCriterionInput
                - studios: HierarchicalMultiCriterionInput
                - tags: HierarchicalMultiCriterionInput

        Returns:
            FindPerformersResultType containing:
                - count: Total number of matching performers
                - performers: List of Performer objects

        Examples:
            Find all favorite performers:
            ```python
            result = await client.find_performers(
                performer_filter={"favorite": True}
            )
            print(f"Found {result.count} favorite performers")
            for performer in result.performers:
                print(f"- {performer.name}")
            ```

            Find performers with specific tags:
            ```python
            result = await client.find_performers(
                performer_filter={
                    "tags": {
                        "value": ["tag1", "tag2"],
                        "modifier": "INCLUDES_ALL"
                    }
                }
            )
            ```

            Find performers with high rating and sort by name:
            ```python
            result = await client.find_performers(
                filter_={
                    "direction": "ASC",
                    "sort": "name",
                },
                performer_filter={
                    "rating100": {
                        "value": 80,
                        "modifier": "GREATER_THAN"
                    }
                }
            )
            ```

            Paginate results:
            ```python
            result = await client.find_performers(
                filter_={
                    "page": 1,
                    "per_page": 25,
                }
            )
            ```
        """
        if filter_ is None:
            filter_ = {"per_page": -1}
        # Add q to filter if provided
        if q is not None:
            filter_ = dict(filter_)  # Copy since we have a default
            filter_["q"] = q
        filter_ = self._normalize_sort_direction(filter_)

        try:
            result = await self.execute(
                fragments.FIND_PERFORMERS_QUERY,
                {"filter": filter_, "performer_filter": performer_filter},
            )
            return FindPerformersResultType(**result["findPerformers"])
        except Exception as e:
            self.log.error(f"Failed to find performers: {e}")
            return FindPerformersResultType(count=0, performers=[])

    async def create_performer(self, performer: Performer) -> Performer:
        """Create a new performer in Stash.

        Args:
            performer: Performer object with the data to create. Required fields:
                - name: Performer name

        Returns:
            Created Performer object with ID and any server-generated fields

        Raises:
            ValueError: If the performer data is invalid
            gql.TransportError: If the request fails

        Examples:
            Create a basic performer:
            ```python
            performer = Performer(
                name="Performer Name",
            )
            created = await client.create_performer(performer)
            print(f"Created performer with ID: {created.id}")
            ```

            Create performer with metadata:
            ```python
            performer = Performer(
                name="Performer Name",
                # Add metadata
                gender="FEMALE",
                birthdate="1990-01-01",
                ethnicity="Caucasian",
                country="USA",
                eye_color="Blue",
                height_cm=170,
                measurements="34B-24-36",
                fake_tits="No",
                career_length="2010-2020",
                tattoos="None",
                piercings="Ears",
                url="https://example.com/performer",
                twitter="@performer",
                instagram="@performer",
                details="Performer details",
            )
            created = await client.create_performer(performer)
            ```

            Create performer with relationships:
            ```python
            performer = Performer(
                name="Performer Name",
                # Add relationships
                tags=[tag1, tag2],
                image="https://example.com/image.jpg",
                stash_ids=[stash_id1, stash_id2],
            )
            created = await client.create_performer(performer)
            ```
        """
        try:
            input_data = await performer.to_input()
            result = await self.execute(
                fragments.CREATE_PERFORMER_MUTATION,
                {"input": input_data},
            )
            return self._decode_result(Performer, result["performerCreate"])
        except Exception as e:
            error_message = str(e)
            if (
                "performer with name" in error_message
                and "already exists" in error_message
            ):
                self.log.info(
                    f"Performer '{performer.name}' already exists. Fetching existing performer."
                )
                # Clear both performer caches since we have a new performer
                # Try to find the existing performer with exact name match
                find_result = await self.find_performers(
                    performer_filter={
                        "name": {"value": performer.name, "modifier": "EQUALS"}
                    },
                )
                if (
                    is_set(find_result.count)
                    and find_result.count > 0
                    and is_set(find_result.performers)
                ):
                    return find_result.performers[0]
                raise  # Re-raise if we couldn't find the performer

            self.log.error(f"Failed to create performer: {e}")
            raise

    async def update_performer(self, performer: Performer) -> Performer:
        """Update an existing performer in Stash.

        Args:
            performer: Performer object with updated data. Required fields:
                - id: Performer ID to update
                Any other fields that are set will be updated.
                Fields that are None will be ignored.

        Returns:
            Updated Performer object with any server-generated fields

        Raises:
            ValueError: If the performer data is invalid
            gql.TransportError: If the request fails

        Examples:
            Update performer name and metadata:
            ```python
            performer = await client.find_performer("123")
            if performer:
                performer.name = "New Name"
                performer.gender = "FEMALE"
                performer.birthdate = "1990-01-01"
                updated = await client.update_performer(performer)
                print(f"Updated performer: {updated.name}")
            ```

            Update performer relationships:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Add new tags
                performer.tags.extend([new_tag1, new_tag2])
                # Update image
                performer.image = "https://example.com/new-image.jpg"
                updated = await client.update_performer(performer)
            ```

            Update performer URLs:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Replace URLs
                performer.url = "https://example.com/new-url"
                performer.twitter = "@new_twitter"
                performer.instagram = "@new_instagram"
                updated = await client.update_performer(performer)
            ```

            Remove performer relationships:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Clear tags
                performer.tags = []
                # Clear image
                performer.image = None
                updated = await client.update_performer(performer)
            ```
        """
        try:
            input_data = await performer.to_input()
            result = await self.execute(
                fragments.UPDATE_PERFORMER_MUTATION,
                {"input": input_data},
            )
            return self._decode_result(Performer, result["performerUpdate"])
        except Exception as e:
            self.log.error(f"Failed to update performer: {e}")
            raise

    async def update_performer_image(
        self, performer: Performer, image_url: str
    ) -> Performer:
        """Update a performer's image.

        Args:
            performer: Performer object with at least the ID set
            image_url: URL or data URI of the image to set

        Returns:
            Updated Performer object with the new image

        Raises:
            ValueError: If the performer data is invalid
            gql.TransportError: If the request fails

        Examples:
            Update performer image with a data URI:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Update with data URI
                image_url = "data:image/jpeg;base64,..."
                updated = await client.update_performer_image(performer, image_url)
            ```

            Update performer image from a file:
            ```python
            performer = await client.find_performer("123")
            if performer:
                # Use the performer's update_avatar method
                updated = await performer.update_avatar(client, "/path/to/image.jpg")
            ```
        """
        try:
            # Create a minimal input with just ID and image
            input_data = {"id": performer.id, "image": image_url}

            result = await self.execute(
                fragments.UPDATE_PERFORMER_MUTATION,
                {"input": input_data},
            )
            return self._decode_result(Performer, result["performerUpdate"])
        except Exception as e:
            self.log.error(f"Failed to update performer image: {e}")
            raise

    async def performer_destroy(
        self,
        input_data: PerformerDestroyInput | dict[str, Any],
    ) -> bool:
        """Delete a performer.

        Args:
            input_data: PerformerDestroyInput object or dictionary containing:
                - id: Performer ID to delete (required)

        Returns:
            True if the performer was successfully deleted

        Raises:
            ValueError: If the performer ID is invalid
            gql.TransportError: If the request fails

        Examples:
            Delete a performer:
            ```python
            result = await client.performer_destroy({"id": "123"})
            print(f"Performer deleted: {result}")
            ```

            Using the input type:
            ```python
            from ...types import PerformerDestroyInput

            input_data = PerformerDestroyInput(id="123")
            result = await client.performer_destroy(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, PerformerDestroyInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be PerformerDestroyInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = PerformerDestroyInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.PERFORMER_DESTROY_MUTATION,
                {"input": input_dict},
            )

            return result.get("performerDestroy") is True
        except Exception as e:
            self.log.error(f"Failed to delete performer: {e}")
            raise

    async def performers_destroy(self, ids: list[str]) -> bool:
        """Delete multiple performers.

        Args:
            ids: List of performer IDs to delete

        Returns:
            True if the performers were successfully deleted

        Raises:
            ValueError: If any performer ID is invalid
            gql.TransportError: If the request fails

        Examples:
            Delete multiple performers:
            ```python
            result = await client.performers_destroy(["123", "456", "789"])
            print(f"Performers deleted: {result}")
            ```
        """
        try:
            result = await self.execute(
                fragments.PERFORMERS_DESTROY_MUTATION,
                {"ids": ids},
            )

            return result.get("performersDestroy") is True
        except Exception as e:
            self.log.error(f"Failed to delete performers: {e}")
            raise

    async def all_performers(self) -> list[Performer]:
        """Get all performers."""
        try:
            result = await self.execute(fragments.ALL_PERFORMERS_QUERY, {})
            performers_data = result.get("allPerformers", [])
            return [self._decode_result(Performer, p) for p in performers_data]
        except Exception as e:
            self.log.error(f"Failed to get all performers: {e}")
            return []

    async def bulk_performer_update(
        self,
        input_data: BulkPerformerUpdateInput | dict[str, Any],
    ) -> list[Performer]:
        """Bulk update performers.

        Args:
            input_data: BulkPerformerUpdateInput object or dictionary containing:
                - ids: List of performer IDs to update (optional)
                - And any fields to update (e.g., gender, birthdate, tags, etc.)

        Returns:
            List of updated Performer objects

        Examples:
            Update multiple performers' gender:
            ```python
            performers = await client.bulk_performer_update({
                "ids": ["1", "2", "3"],
                "gender": "FEMALE"
            })
            ```

            Add tags to multiple performers:
            ```python
            from stash_graphql_client.types import BulkPerformerUpdateInput, BulkUpdateIds

            input_data = BulkPerformerUpdateInput(
                ids=["1", "2", "3"],
                tag_ids=BulkUpdateIds(ids=["tag1", "tag2"], mode="ADD")
            )
            performers = await client.bulk_performer_update(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, BulkPerformerUpdateInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be BulkPerformerUpdateInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = BulkPerformerUpdateInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.BULK_PERFORMER_UPDATE_MUTATION,
                {"input": input_dict},
            )

            performers_data = result.get("bulkPerformerUpdate", [])
            return [self._decode_result(Performer, p) for p in performers_data]
        except Exception as e:
            self.log.error(f"Failed to bulk update performers: {e}")
            raise

    async def performer_merge(
        self,
        input_data: PerformerMergeInput | dict[str, Any],
    ) -> Performer:
        """Merge performers into a single performer.

        **Minimum Stash Version**: v0.30.2+ or commit `65e82a0` or newer

        This feature requires Stash v0.30.2 or later (or development builds with
        commit 65e82a0+). Older Stash versions will raise a GraphQL error.

        Args:
            input_data: PerformerMergeInput object or dictionary containing:
                - source: List of performer IDs to merge (required)
                - destination: ID of the performer to merge into (required)
                - values: Optional PerformerUpdateInput to override destination values

        Returns:
            The merged Performer object

        Raises:
            StashGraphQLError: If Stash version is too old or merge fails

        Examples:
            Merge performers:
            ```python
            merged = await client.performer_merge({
                "source": ["performer1-id", "performer2-id"],
                "destination": "destination-performer-id"
            })
            ```

            Merge with overrides:
            ```python
            from stash_graphql_client.types import PerformerMergeInput, PerformerUpdateInput

            merged = await client.performer_merge(
                PerformerMergeInput(
                    source=["performer1-id", "performer2-id"],
                    destination="destination-performer-id",
                    values=PerformerUpdateInput(
                        id="destination-performer-id",
                        name="New Name",
                        gender="FEMALE"
                    )
                )
            )
            ```
        """
        # Convert PerformerMergeInput to dict if needed
        if isinstance(input_data, PerformerMergeInput):
            input_dict = input_data.to_graphql()
        else:
            # Validate dict structure through Pydantic
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be PerformerMergeInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = PerformerMergeInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            return await self.execute(
                fragments.PERFORMER_MERGE_MUTATION,
                {"input": input_dict},
                result_type=Performer,
            )
        except Exception as e:
            # Provide helpful error for older Stash versions
            error_msg = str(e).lower()
            if "performermerge" in error_msg or "unknown" in error_msg:
                raise StashGraphQLError(
                    "performerMerge requires Stash v0.30.2+ (or commit 65e82a0+). "
                    f"Your Stash version may be too old. Original error: {e}"
                ) from e
            raise

    async def map_performer_ids(
        self,
        performers: list[str | dict[str, Any] | Performer],
        create: bool = False,
        on_multiple: OnMultipleMatch = OnMultipleMatch.RETURN_FIRST,
    ) -> list[str]:
        """Convert performer names/objects to IDs, optionally creating missing performers.

        This is a convenience method to resolve performer references to their IDs,
        reducing boilerplate when working with performer relationships.

        Args:
            performers: List of performer references, can be:
                - str: Performer name (searches by name, then aliases)
                - dict: Performer filter criteria (e.g., {"name": "John", "disambiguation": "Actor"})
                - Performer: Performer object (extracts ID if present, otherwise searches by name)
            create: If True, create performers that don't exist. Default is False.
            on_multiple: Strategy when multiple matches found. Default is RETURN_FIRST.

        Returns:
            List of performer IDs. Skips performers that aren't found (unless create=True).

        Examples:
            Map performer names to IDs:
            ```python
            performer_ids = await client.map_performer_ids(["Jane Doe", "John Smith"])
            # Returns: ["1", "2"] (IDs of existing performers)
            ```

            Auto-create missing performers:
            ```python
            performer_ids = await client.map_performer_ids(
                ["Jane Doe", "NewPerformer"],
                create=True
            )
            # Creates "NewPerformer" if it doesn't exist
            ```

            Handle multiple matches:
            ```python
            performer_ids = await client.map_performer_ids(
                ["AmbiguousName"],
                on_multiple=OnMultipleMatch.RETURN_NONE  # Skip ambiguous matches
            )
            ```

            Use in scene creation:
            ```python
            scene = Scene(
                title="My Scene",
                performers=[]  # Will populate with performer IDs
            )
            scene.performers = await client.map_performer_ids(
                ["Jane Doe", "John Smith"],
                create=True
            )
            created_scene = await client.create_scene(scene)
            ```
        """
        performer_ids: list[str] = []

        for performer_input in performers:
            try:
                # Handle Performer objects
                performer_name: str | None = None
                if isinstance(performer_input, Performer):
                    # If performer has a server-assigned ID, use it directly
                    if not performer_input.is_new():
                        performer_ids.append(performer_input.id)
                        continue
                    # Otherwise search by name
                    if is_set(performer_input.name):
                        performer_name = performer_input.name or ""

                # Handle string input (performer name)
                if isinstance(performer_input, str):
                    performer_name = performer_input

                if performer_name:
                    # Search by name first
                    results = await self.find_performers(
                        performer_filter={
                            "name": {"value": performer_name, "modifier": "EQUALS"}
                        }
                    )

                    # If not found by name, try aliases
                    if is_set(results.count) and results.count == 0:
                        results = await self.find_performers(
                            performer_filter={
                                "aliases": {
                                    "value": performer_name,
                                    "modifier": "INCLUDES",
                                }
                            }
                        )

                    if is_set(results.count) and results.count > 1:
                        # Handle multiple matches based on strategy
                        if on_multiple == OnMultipleMatch.RETURN_NONE:
                            self.log.warning(
                                f"Multiple performers matched '{performer_name}', skipping (on_multiple=RETURN_NONE)"
                            )
                            continue
                        if on_multiple == OnMultipleMatch.RETURN_FIRST:
                            self.log.warning(
                                f"Multiple performers matched '{performer_name}', using first match"
                            )
                            if is_set(results.performers):
                                performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]
                        # RETURN_LIST not applicable here since we're building a flat ID list
                        else:
                            self.log.warning(
                                f"Multiple performers matched '{performer_name}', using first match"
                            )
                            if is_set(results.performers):
                                performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]

                    elif is_set(results.count) and results.count == 1:
                        if is_set(results.performers):
                            performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]
                    elif create:
                        # Create new performer
                        self.log.info(f"Creating missing performer: '{performer_name}'")
                        new_performer = await self.create_performer(
                            Performer(name=performer_name)
                        )
                        performer_ids.append(new_performer.id)  # type: ignore[union-attr]
                    else:
                        self.log.warning(
                            f"Performer '{performer_name}' not found and create=False"
                        )

                # Handle dict input (filter criteria)
                elif isinstance(performer_input, dict):
                    name = performer_input.get("name")
                    if name:
                        results = await self.find_performers(
                            performer_filter={
                                "name": {"value": name, "modifier": "EQUALS"}
                            }
                        )

                        if is_set(results.count) and results.count > 1:
                            if on_multiple == OnMultipleMatch.RETURN_NONE:
                                self.log.warning(
                                    f"Multiple performers matched '{name}', skipping (on_multiple=RETURN_NONE)"
                                )
                                continue
                            if on_multiple == OnMultipleMatch.RETURN_FIRST:
                                self.log.warning(
                                    f"Multiple performers matched '{name}', using first match"
                                )
                                if is_set(results.performers):
                                    performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]
                            else:
                                # Default behavior: use first match
                                self.log.warning(
                                    f"Multiple performers matched '{name}', using first match"
                                )
                                if is_set(results.performers):
                                    performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]
                        elif is_set(results.count) and results.count == 1:
                            if is_set(results.performers):
                                performer_ids.append(results.performers[0].id)  # type: ignore[union-attr]
                        elif create:
                            self.log.info(f"Creating missing performer: '{name}'")
                            new_performer = await self.create_performer(
                                Performer(name=name)
                            )
                            performer_ids.append(new_performer.id)  # type: ignore[union-attr]
                        else:
                            self.log.warning(
                                f"Performer '{name}' not found and create=False"
                            )

            except Exception as e:
                self.log.error(f"Failed to map performer {performer_input}: {e}")
                continue

        return performer_ids



"""DLNA types from schema/types/dlna.graphql."""

from __future__ import annotations

from pydantic import BaseModel, Field

from .base import FromGraphQLMixin, StashInput
from .scalars import Time
from .unset import UNSET, UnsetType


class DLNAIP(FromGraphQLMixin, BaseModel):
    """DLNA IP address type from schema/types/dlna.graphql."""

    ip_address: str | UnsetType = Field(default=UNSET, alias="ipAddress")  # String!
    until: Time | None | UnsetType = (
        UNSET  # Time (Time until IP will be no longer allowed/disallowed)
    )


class DLNAStatus(FromGraphQLMixin, BaseModel):
    """DLNA status from schema/types/dlna.graphql."""

    running: bool | UnsetType = UNSET  # Boolean!
    until: Time | None | UnsetType = (
        UNSET  # Time (If not currently running, time until it will be started. If running, time until it will be stopped)
    )
    recent_ip_addresses: list[str] | UnsetType = Field(
        default=UNSET, alias="recentIPAddresses"
    )  # [String!]!
    allowed_ip_addresses: list[DLNAIP] | UnsetType = Field(
        default=UNSET, alias="allowedIPAddresses"
    )  # [DLNAIP!]!


class EnableDLNAInput(StashInput):
    """Input for enabling DLNA."""

    duration: int | None | UnsetType = (
        UNSET  # Int (Duration to enable, in minutes. 0 or null for indefinite.)
    )


class DisableDLNAInput(StashInput):
    """Input for disabling DLNA."""

    duration: int | None | UnsetType = (
        UNSET  # Int (Duration to enable, in minutes. 0 or null for indefinite.)
    )


class AddTempDLNAIPInput(StashInput):
    """Input for adding temporary DLNA IP."""

    address: str | UnsetType = UNSET  # String!
    duration: int | None | UnsetType = (
        UNSET  # Int (Duration to enable, in minutes. 0 or null for indefinite.)
    )


class RemoveTempDLNAIPInput(StashInput):
    """Input for removing temporary DLNA IP."""

    address: str | UnsetType = UNSET  # String!



"""Group types from schema/types/group.graphql."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from .base import (
    BulkUpdateIds,
    BulkUpdateStrings,
    RelationshipMetadata,
    StashInput,
    StashObject,
    StashResult,
)
from .enums import BulkUpdateIdMode
from .unset import UNSET, UnsetType


if TYPE_CHECKING:
    from .scene import Scene
    from .studio import Studio
    from .tag import Tag


class GroupDescription(BaseModel):
    """Group description type from schema."""

    group: Group | UnsetType = UNSET  # Group!
    description: str | None | UnsetType = UNSET  # String


class GroupCreateInput(StashInput):
    """Input for creating groups from schema/types/group.graphql."""

    # Required fields
    name: str | UnsetType = UNSET  # String!

    # Optional fields
    aliases: str | None | UnsetType = UNSET  # String
    duration: int | None | UnsetType = UNSET  # Int (in seconds)
    date: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = Field(
        default=UNSET, ge=0, le=100
    )  # Int (1-100)
    studio_id: str | None | UnsetType = UNSET  # ID
    director: str | None | UnsetType = UNSET  # String
    synopsis: str | None | UnsetType = UNSET  # String
    urls: list[str] | None | UnsetType = UNSET  # [String!]
    tag_ids: list[str] | None | UnsetType = UNSET  # [ID!]
    containing_groups: list[GroupDescriptionInput] | None | UnsetType = (
        UNSET  # [GroupDescriptionInput!]
    )
    sub_groups: list[GroupDescriptionInput] | None | UnsetType = (
        UNSET  # [GroupDescriptionInput!]
    )
    front_image: str | None | UnsetType = (
        UNSET  # String (URL or base64 encoded data URL)
    )
    back_image: str | None | UnsetType = UNSET  # String (URL or base64)


class GroupUpdateInput(StashInput):
    """Input for updating groups from schema/types/group.graphql."""

    # Required fields
    id: str  # ID!

    # Optional fields
    name: str | None | UnsetType = UNSET  # String
    aliases: str | None | UnsetType = UNSET  # String
    duration: int | None | UnsetType = UNSET  # Int (in seconds)
    date: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = Field(
        default=UNSET, ge=0, le=100
    )  # Int (1-100)
    studio_id: str | None | UnsetType = UNSET  # ID
    director: str | None | UnsetType = UNSET  # String
    synopsis: str | None | UnsetType = UNSET  # String
    urls: list[str] | None | UnsetType = UNSET  # [String!]
    tag_ids: list[str] | None | UnsetType = UNSET  # [ID!]
    containing_groups: list[GroupDescriptionInput] | None | UnsetType = (
        UNSET  # [GroupDescriptionInput!]
    )
    sub_groups: list[GroupDescriptionInput] | None | UnsetType = (
        UNSET  # [GroupDescriptionInput!]
    )
    front_image: str | None | UnsetType = (
        UNSET  # String (URL or base64 encoded data URL)
    )
    back_image: str | None | UnsetType = (
        UNSET  # String (URL or base64 encoded data URL)
    )


class Group(StashObject):
    """Group type from schema."""

    __type_name__ = "Group"
    __update_input_type__ = GroupUpdateInput
    __create_input_type__ = GroupCreateInput

    # Fields to track for changes - only fields that can be written via input types
    __tracked_fields__ = {
        "name",  # GroupCreateInput/GroupUpdateInput
        "urls",  # GroupCreateInput/GroupUpdateInput
        "tags",  # mapped to tag_ids
        "containing_groups",  # GroupCreateInput/GroupUpdateInput
        "sub_groups",  # GroupCreateInput/GroupUpdateInput
        "aliases",  # GroupCreateInput/GroupUpdateInput
        "duration",  # GroupCreateInput/GroupUpdateInput
        "date",  # GroupCreateInput/GroupUpdateInput
        "studio",  # mapped to studio_id
        "director",  # GroupCreateInput/GroupUpdateInput
        "synopsis",  # GroupCreateInput/GroupUpdateInput
    }

    # Required fields
    name: str | UnsetType = UNSET  # String!
    urls: list[str] | UnsetType = Field(default=UNSET)  # [String!]!
    tags: list[Tag] | UnsetType = Field(default=UNSET)  # [Tag!]!
    containing_groups: list[GroupDescription] | UnsetType = Field(
        default=UNSET
    )  # [GroupDescription!]!
    sub_groups: list[GroupDescription] | UnsetType = Field(
        default=UNSET
    )  # [GroupDescription!]!
    scenes: list[Scene] | UnsetType = Field(default=UNSET)  # [Scene!]!

    # Optional fields
    aliases: str | None | UnsetType = UNSET  # String
    duration: int | None | UnsetType = UNSET  # Int (in seconds)
    date: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = Field(
        default=UNSET, ge=0, le=100
    )  # Int (0-100)
    studio: Studio | None | UnsetType = UNSET  # Studio
    director: str | None | UnsetType = UNSET  # String
    synopsis: str | None | UnsetType = UNSET  # String
    front_image_path: str | None | UnsetType = UNSET  # String (Resolver)
    back_image_path: str | None | UnsetType = UNSET  # String (Resolver)
    scene_count: int | None | UnsetType = Field(
        default=UNSET, ge=0
    )  # Int! (Resolver with optional depth)
    performer_count: int | None | UnsetType = Field(
        default=UNSET, ge=0
    )  # Int! (Resolver with optional depth)
    sub_group_count: int | None | UnsetType = Field(
        default=UNSET, ge=0
    )  # Int! (Resolver with optional depth)
    o_counter: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int (Resolver)

    # Field definitions with their conversion functions
    __field_conversions__ = {
        "name": str,
        "urls": list,
        "aliases": str,
        "duration": int,
        "date": str,
        "rating100": int,
        "director": str,
        "synopsis": str,
    }

    __relationships__ = {
        "studio": RelationshipMetadata(
            target_field="studio_id",
            is_list=False,
            query_field="studio",
            inverse_type="Studio",
            query_strategy="filter_query",
            filter_query_hint="findGroups(group_filter={studios: {value: [studio_id]}})",
            notes="Studio has group_count and filter queries, not direct groups field",
        ),
        "tags": RelationshipMetadata(
            target_field="tag_ids",
            is_list=True,
            query_field="tags",
            inverse_type="Tag",
            query_strategy="direct_field",
            notes="Tag has group_count resolver, not direct groups list",
        ),
        # Pattern C: Complex objects with metadata
        "containing_groups": RelationshipMetadata(
            target_field="containing_groups",
            is_list=True,
            transform=lambda g: GroupDescriptionInput(
                group_id=g.group.id if hasattr(g, "group") else g.id,
                description=g.description if hasattr(g, "description") else None,
            ),
            query_field="containing_groups",
            inverse_type="Group",
            inverse_query_field="sub_groups",
            query_strategy="complex_object",
            notes="Uses GroupDescription wrapper with nested group + description metadata",
        ),
        "sub_groups": RelationshipMetadata(
            target_field="sub_groups",
            is_list=True,
            transform=lambda g: GroupDescriptionInput(
                group_id=g.group.id if hasattr(g, "group") else g.id,
                description=g.description if hasattr(g, "description") else None,
            ),
            query_field="sub_groups",
            inverse_type="Group",
            inverse_query_field="containing_groups",
            query_strategy="complex_object",
            notes="Uses GroupDescription wrapper with nested group + description metadata",
        ),
    }

    async def add_sub_group(
        self, sub_group: Group | GroupDescription, description: str | None = None
    ) -> None:
        """Add sub-group (syncs inverse automatically, call save() to persist).

        Args:
            sub_group: Either a Group object or a GroupDescription object
            description: Optional description for the relationship (only used if sub_group is a Group)
        """
        from stash_graphql_client.errors import StashIntegrationError

        # Get current sub_groups value
        current = self.sub_groups

        # If UNSET, fetch from store
        if isinstance(current, UnsetType):
            if self._store is None:
                raise StashIntegrationError(
                    "Cannot add sub-group: store not available. "
                    f"Ensure the {self.__type_name__} was loaded through a StashEntityStore."
                )
            await self._store.populate(self, fields=["sub_groups"])
            current = self.sub_groups

        # Initialize to [] if None or still UNSET after populate
        if current is None or isinstance(current, UnsetType):
            current = []

        # Wrap plain Group in GroupDescription
        if isinstance(sub_group, Group):
            sub_group_desc = GroupDescription(group=sub_group, description=description)
        else:
            sub_group_desc = sub_group

        # Deduplicate by group ID
        group_id = (
            sub_group_desc.group.id if hasattr(sub_group_desc.group, "id") else None
        )
        if group_id:
            # Remove existing entry with same group ID
            current = [
                sg
                for sg in current
                if not (
                    hasattr(sg, "group")
                    and hasattr(sg.group, "id")
                    and sg.group.id == group_id
                )
            ]

        # Add the new entry
        current.append(sub_group_desc)

        # Update the field
        self.sub_groups = current

    async def remove_sub_group(self, sub_group: Group | GroupDescription) -> None:
        """Remove sub-group (syncs inverse automatically, call save() to persist).

        Args:
            sub_group: Either a Group object or GroupDescription object to remove
        """
        current = self.sub_groups

        # Skip if UNSET or None
        if isinstance(current, UnsetType) or current is None:
            return

        # Extract group ID to match against
        if isinstance(sub_group, Group):
            group_id = sub_group.id if hasattr(sub_group, "id") else None
        else:  # GroupDescription
            group_id = sub_group.group.id if hasattr(sub_group.group, "id") else None

        if group_id:
            # Remove by matching group ID
            self.sub_groups = [
                sg
                for sg in current
                if not (
                    hasattr(sg, "group")
                    and hasattr(sg.group, "id")
                    and sg.group.id == group_id
                )
            ]

    async def add_containing_group(
        self, containing_group: Group | GroupDescription
    ) -> None:
        """Add containing group (syncs inverse automatically, call save() to persist).

        Args:
            containing_group: Either a Group object (will be wrapped with None description)
                             or a GroupDescription object (used as-is)
        """
        from stash_graphql_client.errors import StashIntegrationError

        # Get current containing_groups value
        current = self.containing_groups

        # If UNSET, fetch from store
        if isinstance(current, UnsetType):
            if self._store is None:
                raise StashIntegrationError(
                    "Cannot add containing group: store not available. "
                    f"Ensure the {self.__type_name__} was loaded through a StashEntityStore."
                )
            await self._store.populate(self, fields=["containing_groups"])
            current = self.containing_groups

        # Initialize to [] if None or still UNSET after populate
        if current is None or isinstance(current, UnsetType):
            current = []

        # Wrap plain Group in GroupDescription
        if isinstance(containing_group, Group):
            containing_group_desc = GroupDescription(
                group=containing_group, description=None
            )
        else:
            containing_group_desc = containing_group

        # Deduplicate by group ID
        group_id = (
            containing_group_desc.group.id
            if hasattr(containing_group_desc.group, "id")
            else None
        )
        if group_id:
            # Remove existing entry with same group ID
            current = [
                cg
                for cg in current
                if not (
                    hasattr(cg, "group")
                    and hasattr(cg.group, "id")
                    and cg.group.id == group_id
                )
            ]

        # Add the new entry
        current.append(containing_group_desc)

        # Update the field
        self.containing_groups = current

    async def remove_containing_group(
        self, containing_group: Group | GroupDescription
    ) -> None:
        """Remove containing group (syncs inverse automatically, call save() to persist).

        Args:
            containing_group: Either a Group object or GroupDescription object to remove
        """
        current = self.containing_groups

        # Skip if UNSET or None
        if isinstance(current, UnsetType) or current is None:
            return

        # Extract group ID to match against
        if isinstance(containing_group, Group):
            group_id = containing_group.id if hasattr(containing_group, "id") else None
        else:  # GroupDescription
            group_id = (
                containing_group.group.id
                if hasattr(containing_group.group, "id")
                else None
            )

        if group_id:
            # Remove by matching group ID
            self.containing_groups = [
                cg
                for cg in current
                if not (
                    hasattr(cg, "group")
                    and hasattr(cg.group, "id")
                    and cg.group.id == group_id
                )
            ]


class GroupDescriptionInput(StashInput):
    """Input for group description."""

    group_id: str | UnsetType = UNSET  # ID!
    description: str | None | UnsetType = UNSET  # String


class BulkUpdateGroupDescriptionsInput(StashInput):
    """Input for bulk updating group descriptions."""

    groups: list[GroupDescriptionInput] | UnsetType = UNSET  # [GroupDescriptionInput!]!
    mode: BulkUpdateIdMode | UnsetType = UNSET  # BulkUpdateIdMode!


class BulkGroupUpdateInput(StashInput):
    """Input for bulk updating groups."""

    client_mutation_id: str | None | UnsetType = Field(
        default=UNSET, alias="clientMutationId"
    )  # String
    ids: list[str] | UnsetType = UNSET  # [ID!]
    rating100: int | None | UnsetType = Field(default=UNSET, ge=0, le=100)  # Int
    studio_id: str | None | UnsetType = UNSET  # ID
    director: str | None | UnsetType = UNSET  # String
    urls: BulkUpdateStrings | None | UnsetType = UNSET  # BulkUpdateStrings
    tag_ids: BulkUpdateIds | None | UnsetType = UNSET  # BulkUpdateIds
    containing_groups: BulkUpdateGroupDescriptionsInput | None | UnsetType = (
        UNSET  # BulkUpdateGroupDescriptionsInput
    )
    sub_groups: BulkUpdateGroupDescriptionsInput | None | UnsetType = (
        UNSET  # BulkUpdateGroupDescriptionsInput
    )


class GroupDestroyInput(StashInput):
    """Input for destroying groups."""

    id: str  # ID!


class ReorderSubGroupsInput(StashInput):
    """Input for reordering sub groups from schema/types/group.graphql.

    Fields:
        group_id: str of the group to reorder sub groups for
        sub_group_ids: strs of the sub groups to reorder. These must be a subset of the current sub groups.
            Sub groups will be inserted in this order at the insert_index.
        insert_at_id: The sub-group ID at which to insert the sub groups
        insert_after: If true, the sub groups will be inserted after the insert_index,
            otherwise they will be inserted before"""

    group_id: str | UnsetType = UNSET  # ID!
    sub_group_ids: list[str] | UnsetType = UNSET  # [ID!]!
    insert_at_id: str | UnsetType = UNSET  # ID!
    insert_after: bool | UnsetType = UNSET  # Boolean


class FindGroupsResultType(StashResult):
    """Result type for finding groups."""

    count: int | UnsetType = UNSET  # Int!
    groups: list[Group] | UnsetType = UNSET  # [Group!]!


class GroupSubGroupAddInput(StashInput):
    """Input for adding sub groups from schema/types/group.graphql.

    Fields:
        containing_group_id: str of the group to add sub groups to
        sub_groups: List of sub groups to add
        insert_index: The index at which to insert the sub groups. If not provided,
            the sub groups will be appended to the end"""

    containing_group_id: str | UnsetType = UNSET  # ID!
    sub_groups: list[GroupDescriptionInput] | UnsetType = (
        UNSET  # [GroupDescriptionInput!]!
    )
    insert_index: int | None | UnsetType = UNSET  # Int


class GroupSubGroupRemoveInput(StashInput):
    """Input for removing sub groups."""

    containing_group_id: str | UnsetType = UNSET  # ID!
    sub_group_ids: list[str] | UnsetType = UNSET  # [ID!]!



"""Marker-related client functionality."""

from typing import Any

from ... import fragments
from ...types import (
    BulkSceneMarkerUpdateInput,
    FindSceneMarkersResultType,
    MarkerStringsResultType,
    SceneMarker,
)
from ..protocols import StashClientProtocol


class MarkerClientMixin(StashClientProtocol):
    """Mixin for marker-related client methods."""

    async def find_marker(self, id: str) -> SceneMarker | None:
        """Find a scene marker by its ID.

        Args:
            id: The ID of the marker to find

        Returns:
            SceneMarker object if found, None otherwise
        """
        try:
            result = await self.execute(
                fragments.FIND_MARKER_QUERY,
                {"id": id},
            )
            if result and result.get("findSceneMarkers"):
                markers = result["findSceneMarkers"].get("scene_markers", [])
                if markers:
                    return self._decode_result(SceneMarker, markers[0])
            return None
        except Exception as e:
            self.log.error(f"Failed to find marker {id}: {e}")
            return None

    async def find_markers(
        self,
        filter_: dict[str, Any] | None = None,
        marker_filter: dict[str, Any] | None = None,
        q: str | None = None,
    ) -> FindSceneMarkersResultType:
        """Find scene markers matching the given filters.

        Args:
            filter_: Optional general filter parameters:
                - q: str (search query)
                - direction: SortDirectionEnum (ASC/DESC)
                - page: int
                - per_page: int
                - sort: str (field to sort by)
            marker_filter: Optional marker-specific filter
            q: Optional search query (alternative to filter_["q"])

        Returns:
            FindSceneMarkersResultType containing:
                - count: Total number of matching markers
                - scene_markers: List of SceneMarker objects
        """
        if filter_ is None:
            filter_ = {"per_page": -1}
        # Add q to filter if provided
        if q is not None:
            filter_ = dict(filter_ or {})
            filter_["q"] = q
        filter_ = self._normalize_sort_direction(filter_)

        try:
            result = await self.execute(
                fragments.FIND_MARKERS_QUERY,
                {"filter": filter_, "marker_filter": marker_filter},
            )
            if result and result.get("findSceneMarkers"):
                return self._decode_result(
                    FindSceneMarkersResultType, result["findSceneMarkers"]
                )
            return FindSceneMarkersResultType(count=0, scene_markers=[])
        except Exception as e:
            self.log.error(f"Failed to find markers: {e}")
            return FindSceneMarkersResultType(count=0, scene_markers=[])

    async def create_marker(self, marker: SceneMarker) -> SceneMarker:
        """Create a new scene marker in Stash.

        Args:
            marker: SceneMarker object with the data to create. Required fields:
                - title: Marker title
                - scene_id: ID of the scene this marker belongs to
                - seconds: Time in seconds where the marker occurs

        Returns:
            Created SceneMarker object with ID and any server-generated fields

        Raises:
            ValueError: If the marker data is invalid
            gql.TransportError: If the request fails
        """
        try:
            input_data = await marker.to_input()
            result = await self.execute(
                fragments.CREATE_MARKER_MUTATION,
                {"input": input_data},
            )
            return self._decode_result(SceneMarker, result["sceneMarkerCreate"])
        except Exception as e:
            self.log.error(f"Failed to create marker: {e}")
            raise

    async def scene_marker_tags(self, scene_id: str) -> list[dict[str, Any]]:
        """Get scene marker tags for a scene.

        Args:
            scene_id: Scene ID

        Returns:
            List of scene marker tags, each containing:
                - tag: Tag object
                - scene_markers: List of SceneMarker objects
        """
        try:
            result = await self.execute(
                fragments.SCENE_MARKER_TAG_QUERY,
                {"scene_id": scene_id},
            )
            # Explicitly convert to list of dictionaries
            return [dict(tag) for tag in result["sceneMarkerTags"]]
        except Exception as e:
            self.log.error(f"Failed to get scene marker tags for scene {scene_id}: {e}")
            return []

    async def update_marker(self, marker: SceneMarker) -> SceneMarker:
        """Update an existing scene marker in Stash.

        Args:
            marker: SceneMarker object with updated data. Required fields:
                - id: Marker ID to update
                Any other fields that are set will be updated.
                Fields that are None will be ignored.

        Returns:
            Updated SceneMarker object with any server-generated fields

        Raises:
            ValueError: If the marker data is invalid
            gql.TransportError: If the request fails
        """
        try:
            input_data = await marker.to_input()
            result = await self.execute(
                fragments.UPDATE_MARKER_MUTATION,
                {"input": input_data},
            )
            return self._decode_result(SceneMarker, result["sceneMarkerUpdate"])
        except Exception as e:
            self.log.error(f"Failed to update marker: {e}")
            raise

    async def scene_marker_destroy(self, id: str) -> bool:
        """Delete a scene marker.

        Args:
            id: Scene marker ID to delete

        Returns:
            True if the scene marker was successfully deleted

        Raises:
            ValueError: If the scene marker ID is invalid
            gql.TransportError: If the request fails
        """
        try:
            result = await self.execute(
                fragments.SCENE_MARKER_DESTROY_MUTATION,
                {"id": id},
            )

            return result.get("sceneMarkerDestroy") is True
        except Exception as e:
            self.log.error(f"Failed to delete scene marker: {e}")
            raise

    async def scene_markers_destroy(self, ids: list[str]) -> bool:
        """Delete multiple scene markers.

        Args:
            ids: List of scene marker IDs to delete

        Returns:
            True if the scene markers were successfully deleted

        Raises:
            ValueError: If any scene marker ID is invalid
            gql.TransportError: If the request fails
        """
        try:
            result = await self.execute(
                fragments.SCENE_MARKERS_DESTROY_MUTATION,
                {"ids": ids},
            )

            return result.get("sceneMarkersDestroy") is True
        except Exception as e:
            self.log.error(f"Failed to delete scene markers: {e}")
            raise

    async def bulk_scene_marker_update(
        self,
        input_data: BulkSceneMarkerUpdateInput | dict[str, Any],
    ) -> list[SceneMarker]:
        """Bulk update scene markers.

        Args:
            input_data: BulkSceneMarkerUpdateInput object or dictionary containing:
                - ids: List of scene marker IDs to update (optional)
                - And any fields to update (e.g., primary_tag_id, tag_ids, etc.)

        Returns:
            List of updated SceneMarker objects

        Examples:
            Update multiple markers' primary tag:
            ```python
            markers = await client.bulk_scene_marker_update({
                "ids": ["1", "2", "3"],
                "primary_tag_id": "tag123"
            })
            ```

            Add tags to multiple markers:
            ```python
            from stash_graphql_client.types import BulkSceneMarkerUpdateInput, BulkUpdateIds

            input_data = BulkSceneMarkerUpdateInput(
                ids=["1", "2", "3"],
                tag_ids=BulkUpdateIds(ids=["tag1", "tag2"], mode="ADD")
            )
            markers = await client.bulk_scene_marker_update(input_data)
            ```
        """
        try:
            # Convert BulkSceneMarkerUpdateInput to dict if needed
            if isinstance(input_data, BulkSceneMarkerUpdateInput):
                input_dict = input_data.to_graphql()
            else:
                # Validate dict structure through Pydantic
                if not isinstance(input_data, dict):
                    raise TypeError(
                        f"input_data must be BulkSceneMarkerUpdateInput or dict, "
                        f"got {type(input_data).__name__}"
                    )
                validated = BulkSceneMarkerUpdateInput(**input_data)
                input_dict = validated.to_graphql()

            result = await self.execute(
                fragments.BULK_SCENE_MARKER_UPDATE_MUTATION,
                {"input": input_dict},
            )

            markers_data = result.get("bulkSceneMarkerUpdate", [])
            return [self._decode_result(SceneMarker, m) for m in markers_data]
        except Exception as e:
            self.log.error(f"Failed to bulk update scene markers: {e}")
            raise

    async def marker_wall(self, q: str | None = None) -> list[SceneMarker]:
        """Get marker wall - random markers for display.

        Args:
            q: Optional search query to filter markers

        Returns:
            List of SceneMarker objects selected for wall display

        Examples:
            Get all wall markers:
            ```python
            markers = await client.marker_wall()
            print(f"Found {len(markers)} markers for wall")
            ```

            Search wall markers:
            ```python
            markers = await client.marker_wall(q="interview")
            ```
        """
        try:
            result = await self.execute(
                fragments.MARKER_WALL_QUERY,
                {"q": q} if q else {},
            )
            markers_data = result.get("markerWall", [])
            return [self._decode_result(SceneMarker, m) for m in markers_data]
        except Exception as e:
            self.log.error(f"Failed to get marker wall: {e}")
            return []

    async def marker_strings(
        self, q: str | None = None, sort: str | None = None
    ) -> list[MarkerStringsResultType]:
        """Get marker title strings with counts.

        Args:
            q: Optional search query to filter marker titles
            sort: Optional sort order for results

        Returns:
            List of MarkerStringsResultType objects containing:
                - id: Marker title ID
                - title: Marker title string
                - count: Number of markers with this title

        Examples:
            Get all marker title strings:
            ```python
            strings = await client.marker_strings()
            for s in strings:
                print(f"{s.title}: {s.count} markers")
            ```

            Search and sort marker strings:
            ```python
            strings = await client.marker_strings(q="bj", sort="count")
            ```
        """
        try:
            result = await self.execute(
                fragments.MARKER_STRINGS_QUERY,
                {"q": q, "sort": sort},
            )
            strings_data = result.get("markerStrings", [])
            return [
                self._decode_result(MarkerStringsResultType, s) for s in strings_data
            ]
        except Exception as e:
            self.log.error(f"Failed to get marker strings: {e}")
            return []



"""Subscription-related client functionality."""

import asyncio
from collections.abc import AsyncIterator, Callable
from contextlib import asynccontextmanager
from typing import Any, TypeVar

from gql import gql

from ...types import JobStatus, JobStatusUpdate, LogEntry
from ...types.unset import UnsetType
from .protocols import StashClientProtocol


T = TypeVar("T")
U = TypeVar("U")


class AsyncIteratorWrapper(AsyncIterator[T]):
    """Wrapper for async iterators that transforms their output."""

    def __init__(self, iterator: AsyncIterator[U], transform: Callable[[U], T]):
        """Initialize the wrapper.

        Args:
            iterator: The async iterator to wrap
            transform: Function to transform the iterator's output
        """
        self.iterator = iterator
        self.transform = transform

    def __aiter__(self):
        """Return self as an async iterator."""
        return self

    async def __anext__(self) -> T:
        """Get the next transformed value from the iterator.

        Returns:
            The next transformed value

        Raises:
            StopAsyncIteration: When the iterator is exhausted
        """
        value = await self.iterator.__anext__()
        return self.transform(value)


class SubscriptionClientMixin(StashClientProtocol):
    """Mixin for subscription-related client methods."""

    @asynccontextmanager
    async def _subscription_client(self):
        """Get a session configured for subscriptions via dedicated WebSocket transport."""
        if not hasattr(self, "_ws_session") or self._ws_session is None:
            raise Exception("Failed to connect")

        # Use the already-connected WebSocket session
        yield self._ws_session

    @asynccontextmanager
    async def subscribe_to_jobs(self) -> AsyncIterator[AsyncIterator[JobStatusUpdate]]:
        """Subscribe to job status updates.

        Returns:
            An async context manager that yields an async iterator of JobStatusUpdate objects

        Example:
            ```python
            async with client.subscribe_to_jobs() as subscription:
                async for update in subscription:
                    print(f"Job {update.job.id}: {update.status} ({update.progress}%)")
                    if update.status == "FINISHED":
                        break
            ```
        """
        subscription = gql(
            """
            subscription {
                jobsSubscribe {
                    type
                    job {
                        id
                        addTime
                        status
                        subTasks
                        description
                        progress
                        error
                    }
                }
            }
        """
        )
        async with self._subscription_client() as session:
            subscription_gen = session.subscribe(subscription)
            yield AsyncIteratorWrapper(
                subscription_gen,
                lambda x: JobStatusUpdate.model_validate(
                    {
                        "type": x["jobsSubscribe"]["type"],
                        "job": x["jobsSubscribe"]["job"],
                    }
                ),
            )

    @asynccontextmanager
    async def subscribe_to_logs(self) -> AsyncIterator[AsyncIterator[list[LogEntry]]]:
        """Subscribe to log entries.

        Returns:
            An async context manager that yields an async iterator of LogEntry lists

        Example:
            ```python
            async with client.subscribe_to_logs() as subscription:
                async for logs in subscription:
                    for entry in logs:
                        print(f"{entry.time} [{entry.level}] {entry.message}")
            ```
        """
        subscription = gql(
            """
            subscription {
                loggingSubscribe {
                    time
                    level
                    message
                }
            }
        """
        )
        async with self._subscription_client() as session:
            subscription_gen = session.subscribe(subscription)
            yield AsyncIteratorWrapper(
                subscription_gen,
                lambda x: [
                    LogEntry.model_validate(entry) for entry in x["loggingSubscribe"]
                ],
            )

    @asynccontextmanager
    async def subscribe_to_scan_complete(self) -> AsyncIterator[AsyncIterator[bool]]:
        """Subscribe to scan completion events.

        Returns:
            An async context manager that yields an async iterator of scan completion events

        Example:
            ```python
            async with client.subscribe_to_scan_complete() as subscription:
                async for _ in subscription:
                    print("Scan completed!")
                    await client.metadata_generate(...)  # Generate after scan
            ```
        """
        subscription = gql(
            """
            subscription {
                scanCompleteSubscribe
            }
        """
        )
        async with self._subscription_client() as session:
            subscription_gen = session.subscribe(subscription)
            yield AsyncIteratorWrapper(
                subscription_gen, lambda x: x["scanCompleteSubscribe"]
            )

    async def _check_job_status(
        self, job_id: str
    ) -> tuple[bool | None, JobStatus | None]:
        """Check current job status.

        Args:
            job_id: Job ID to check

        Returns:
            Tuple of (is_done, status):
            - is_done: True if job is in final state, False if running, None if not found
            - status: Current job status if found, None if not found
        """
        result = await self.execute(
            """
            query FindJob($id: ID!) {
                findJob(input: {id: $id}) {
                    id
                    status
                    progress
                    description
                }
            }
            """,
            {"id": job_id},
        )

        job: dict[str, Any] | None = result.get("findJob")
        if job is not None:
            job_status = JobStatus(job["status"])
            progress = job.get("progress") or 0
            description = job.get("description") or ""
            self.log.info(
                f"Job {job_id}: {job_status} ({progress:.1f}%) - {description}"
            )

            is_done = job_status in [JobStatus.FINISHED, JobStatus.CANCELLED]
            return is_done, job_status

        return None, None

    async def wait_for_job_with_updates(
        self,
        job_id: str,
        status: JobStatus = JobStatus.FINISHED,
        timeout: float = 120,
    ) -> bool | None:
        """Wait for a job to complete with real-time updates.

        Args:
            job_id: Job ID to wait for
            status: Status to wait for
            timeout: Maximum time to wait in seconds

        Returns:
            True if job reached desired status
            False if job finished with different status
            None if job not found

        Example:
            ```python
            job_id = await client.metadata_generate(...)
            if await client.wait_for_job_with_updates(job_id):
                print("Generation complete!")
            ```
        """
        try:
            # First check if the job is already finished
            is_done, job_status = await self._check_job_status(job_id)
            if is_done is None:
                return None  # Job not found
            if is_done:
                return job_status == status

            # Job not finished, wait for updates
            async with asyncio.timeout(timeout):
                async with self.subscribe_to_jobs() as subscription:
                    async for update in subscription:
                        # update.job is always populated (Job! in GraphQL schema)
                        job_data = update.job

                        # Check if this is the job we're waiting for
                        if job_data.id != job_id:
                            continue

                        # Extract status and other fields from the Job model
                        job_status_value = job_data.status
                        job_progress = (
                            job_data.progress
                            if not isinstance(job_data.progress, UnsetType)
                            else 0
                        )
                        job_description = (
                            job_data.description
                            if not isinstance(job_data.description, UnsetType)
                            else ""
                        )

                        # Status is always JobStatus enum (Pydantic validates it)
                        if isinstance(job_status_value, JobStatus):
                            job_status = job_status_value
                        else:
                            self.log.warning(
                                "Received job update without valid status for job %s",
                                job_id,
                            )
                            continue

                        # Log update
                        self.log.info(
                            f"Job {job_id}: {job_status.value} "
                            f"({job_progress:.1f}%) - {job_description}"
                        )

                        # Check if we've reached the desired status
                        if job_status == status:
                            return True
                        if job_status in [
                            JobStatus.FINISHED,
                            JobStatus.CANCELLED,
                        ]:
                            return False

            return None
        except TimeoutError:
            self.log.error(f"Timeout waiting for job {job_id}")
            return None
        except Exception as e:
            self.log.error(f"Failed to wait for job {job_id}: {e}")
            # Fall back to polling if subscription fails
            return await self.wait_for_job(job_id, status, timeout=timeout)



"""File-related client functionality."""

from typing import Any

from ... import fragments
from ...types import (
    AssignSceneFileInput,
    BaseFile,
    FileSetFingerprintsInput,
    FindFilesResultType,
    FindFoldersResultType,
    Folder,
    FolderFilterType,
    MoveFilesInput,
)
from ..protocols import StashClientProtocol


class FileClientMixin(StashClientProtocol):
    """Mixin for file-related client methods."""

    async def find_file(
        self,
        id: str | None = None,
        path: str | None = None,
    ) -> BaseFile | None:
        """Find a file by its ID or path.

        Args:
            id: The ID of the file to find (optional)
            path: The path of the file to find (optional)

        Returns:
            BaseFile object (VideoFile, ImageFile, GalleryFile, or BasicFile) if found,
            None otherwise

        Raises:
            ValueError: If neither id nor path is provided

        Examples:
            Find a file by ID:
            ```python
            file = await client.find_file(id="123")
            if file:
                print(f"Found file: {file.path}")
            ```

            Find a file by path:
            ```python
            file = await client.find_file(path="/path/to/video.mp4")
            if file:
                print(f"File size: {file.size} bytes")
            ```

            Check file type:
            ```python
            from stash_graphql_client.types import VideoFile, ImageFile

            file = await client.find_file(path="/path/to/file.mp4")
            if isinstance(file, VideoFile):
                print(f"Video: {file.width}x{file.height}, {file.duration}s")
            elif isinstance(file, ImageFile):
                print(f"Image: {file.width}x{file.height}")
            ```
        """
        if id is None and path is None:
            raise ValueError("Either id or path must be provided")

        try:
            # BaseFile is polymorphic - will return VideoFile, ImageFile, etc
            return await self.execute(
                fragments.FIND_FILE_QUERY,
                {"id": id, "path": path},
                result_type=BaseFile,
            )
        except Exception as e:
            self.log.error(f"Failed to find file (id={id}, path={path}): {e}")
            return None

    async def find_files(
        self,
        file_filter: dict[str, Any] | None = None,
        filter_: dict[str, Any] | None = None,
        ids: list[str] | None = None,
    ) -> FindFilesResultType:
        """Find files matching the given filters.

        Args:
            file_filter: Optional file-specific filter:
                - path: StringCriterionInput
                - basename: StringCriterionInput
                - dir: StringCriterionInput
                - parent_folder: HierarchicalMultiCriterionInput
                - zip_file: MultiCriterionInput
                - mod_time: TimestampCriterionInput
                - size: IntCriterionInput
            filter_: Optional general filter parameters:
                - q: str (search query)
                - direction: SortDirectionEnum (ASC/DESC)
                - page: int
                - per_page: int
                - sort: str (field to sort by)
            ids: Optional list of file IDs to retrieve

        Returns:
            FindFilesResultType containing:
                - count: Total number of matching files
                - megapixels: Total megapixels of image files
                - duration: Total duration in seconds of video files
                - size: Total size in bytes
                - files: List of BaseFile objects

        Examples:
            Find all files in a directory:
            ```python
            result = await client.find_files(
                file_filter={
                    "path": {
                        "value": "/videos/",
                        "modifier": "INCLUDES"
                    }
                }
            )
            print(f"Found {result.count} files, total size: {result.size} bytes")
            ```

            Find video files by size:
            ```python
            result = await client.find_files(
                file_filter={
                    "size": {
                        "value": 1000000000,  # 1GB
                        "modifier": "GREATER_THAN"
                    }
                }
            )
            for file in result.files:
                print(f"{file.path}: {file.size} bytes")
            ```

            Find files by IDs:
            ```python
            result = await client.find_files(ids=["1", "2", "3"])
            ```
        """
        if filter_ is not None:
            filter_ = self._normalize_sort_direction(filter_)
        try:
            return await self.execute(
                fragments.FIND_FILES_QUERY,
                {
                    "file_filter": file_filter,
                    "filter": filter_,
                    "ids": ids,
                },
                result_type=FindFilesResultType,
            )
        except Exception as e:
            self.log.error(f"Failed to find files: {e}")
            return FindFilesResultType(
                count=0,
                megapixels=0.0,
                duration=0.0,
                size=0,
                files=[],
            )

    async def move_files(self, input_data: MoveFilesInput | dict[str, Any]) -> bool:
        """Move files to a new location.

        Args:
            input_data: MoveFilesInput object or dictionary containing:
                - ids: List of file IDs to move (required)
                - destination_folder: Destination folder path (optional)
                - destination_folder_id: Destination folder ID (optional)
                - destination_basename: New basename for single file (optional)

                Note: Either destination_folder or destination_folder_id must be provided

        Returns:
            True if the move was successful, False otherwise

        Examples:
            Move files to a new folder by path:
            ```python
            success = await client.move_files({
                "ids": ["1", "2", "3"],
                "destination_folder": "/new/location"
            })
            ```

            Move files to a new folder by ID:
            ```python
            success = await client.move_files({
                "ids": ["1", "2"],
                "destination_folder_id": "folder123"
            })
            ```

            Move and rename a single file:
            ```python
            success = await client.move_files({
                "ids": ["1"],
                "destination_folder": "/new/location",
                "destination_basename": "newname.mp4"
            })
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, MoveFilesInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be MoveFilesInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = MoveFilesInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.MOVE_FILES_MUTATION,
                {"input": input_dict},
            )
            return result.get("moveFiles") is True
        except Exception as e:
            self.log.error(f"Failed to move files: {e}")
            return False

    async def file_set_fingerprints(
        self,
        input_data: FileSetFingerprintsInput | dict[str, Any],
    ) -> bool:
        """Set fingerprints for a file.

        Args:
            input_data: FileSetFingerprintsInput object or dictionary containing:
                - id: File ID (required)
                - fingerprints: List of SetFingerprintsInput objects with:
                    - type: Fingerprint type (required)
                    - value: Fingerprint value (optional)

        Returns:
            True if the operation was successful, False otherwise

        Examples:
            Set MD5 fingerprint:
            ```python
            success = await client.file_set_fingerprints({
                "id": "file123",
                "fingerprints": [
                    {"type": "MD5", "value": "abc123def456"}
                ]
            })
            ```

            Set multiple fingerprints:
            ```python
            success = await client.file_set_fingerprints({
                "id": "file123",
                "fingerprints": [
                    {"type": "MD5", "value": "abc123"},
                    {"type": "PHASH", "value": "def456"},
                ]
            })
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, FileSetFingerprintsInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be FileSetFingerprintsInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = FileSetFingerprintsInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.FILE_SET_FINGERPRINTS_MUTATION,
                {"input": input_dict},
            )
            return result.get("fileSetFingerprints") is True
        except Exception as e:
            self.log.error(f"Failed to set file fingerprints: {e}")
            return False

    async def scene_assign_file(
        self,
        input_data: AssignSceneFileInput | dict[str, Any],
    ) -> bool:
        """Assign a file to a scene.

        Args:
            input_data: AssignSceneFileInput object or dictionary containing:
                - scene_id: Scene ID (required)
                - file_id: File ID (required)

        Returns:
            True if the assignment was successful, False otherwise

        Examples:
            Assign a file to a scene:
            ```python
            success = await client.scene_assign_file({
                "scene_id": "scene123",
                "file_id": "file456"
            })
            ```

            Using the input type:
            ```python
            from stash_graphql_client.types import AssignSceneFileInput

            input_data = AssignSceneFileInput(
                scene_id="scene123",
                file_id="file456"
            )
            success = await client.scene_assign_file(input_data)
            ```
        """
        # Validate input type before try block so TypeError propagates
        if isinstance(input_data, AssignSceneFileInput):
            input_dict = input_data.to_graphql()
        else:
            if not isinstance(input_data, dict):
                raise TypeError(
                    f"input_data must be AssignSceneFileInput or dict, "
                    f"got {type(input_data).__name__}"
                )
            validated = AssignSceneFileInput(**input_data)
            input_dict = validated.to_graphql()

        try:
            result = await self.execute(
                fragments.SCENE_ASSIGN_FILE_MUTATION,
                {"input": input_dict},
            )
            return result.get("sceneAssignFile") is True
        except Exception as e:
            self.log.error(f"Failed to assign file to scene: {e}")
            return False

    async def delete_files(self, ids: list[str]) -> bool:
        """Delete files.

        Args:
            ids: List of file IDs to delete

        Returns:
            True if the files were successfully deleted

        Raises:
            ValueError: If any file ID is invalid
            gql.TransportError: If the request fails
        """
        try:
            result = await self.execute(
                fragments.DELETE_FILES_MUTATION,
                {"ids": ids},
            )

            return result.get("deleteFiles") is True
        except Exception as e:
            self.log.error(f"Failed to delete files: {e}")
            raise

    async def find_folder(
        self,
        id: str | None = None,
        path: str | None = None,
    ) -> Folder | None:
        """Find a folder by its ID or path.

        Args:
            id: The ID of the folder to find (optional)
            path: The path of the folder to find (optional)

        Returns:
            Folder object if found, None otherwise

        Raises:
            ValueError: If neither id nor path is provided
        """
        if id is None and path is None:
            raise ValueError("Either id or path must be provided")

        try:
            return await self.execute(
                fragments.FIND_FOLDER_QUERY,
                {"id": id, "path": path},
                result_type=Folder,
            )
        except Exception as e:
            self.log.error(f"Failed to find folder (id={id}, path={path}): {e}")
            return None

    async def find_folders(
        self,
        folder_filter: FolderFilterType | dict[str, Any] | None = None,
        filter_: dict[str, Any] | None = None,
        ids: list[str] | None = None,
    ) -> FindFoldersResultType:
        """Find folders matching the given filters.

        Args:
            folder_filter: Optional folder-specific filter (FolderFilterType or dict)
            filter_: Optional general filter parameters (FindFilterType or dict)
            ids: Optional list of folder IDs to retrieve

        Returns:
            FindFoldersResultType containing count and list of folders
        """
        try:
            # Convert FolderFilterType to dict if needed
            folder_filter_dict: dict[str, Any] | None = None
            if isinstance(folder_filter, FolderFilterType):
                folder_filter_dict = folder_filter.to_graphql()
            else:
                folder_filter_dict = folder_filter

            return await self.execute(
                fragments.FIND_FOLDERS_QUERY,
                {
                    "folder_filter": folder_filter_dict,
                    "filter": filter_,
                    "ids": ids,
                },
                result_type=FindFoldersResultType,
            )
        except Exception as e:
            self.log.error(f"Failed to find folders: {e}")
            return FindFoldersResultType(count=0, folders=[])



"""Image types from schema/types/image.graphql."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field, field_validator

from .base import (
    BulkUpdateIds,
    BulkUpdateStrings,
    RelationshipMetadata,
    StashInput,
    StashObject,
    StashResult,
)
from .files import ImageFile, VideoFile
from .unset import UNSET, UnsetType


if TYPE_CHECKING:
    from .gallery import Gallery
    from .performer import Performer
    from .studio import Studio
    from .tag import Tag

# VisualFile union type (VideoFile | ImageFile)
VisualFile = VideoFile | ImageFile


class ImageFileType(BaseModel):
    """Image file type from schema/types/image.graphql."""

    mod_time: datetime | UnsetType = UNSET  # Time!
    size: int | UnsetType = UNSET  # Int!
    width: int | UnsetType = UNSET  # Int!
    height: int | UnsetType = UNSET  # Int!


class ImagePathsType(BaseModel):
    """Image paths type from schema/types/image.graphql."""

    thumbnail: str | None | UnsetType = UNSET  # String (Resolver)
    preview: str | None | UnsetType = UNSET  # String (Resolver)
    image: str | None | UnsetType = UNSET  # String (Resolver)


class ImageUpdateInput(StashInput):
    """Input for updating images."""

    # Required fields
    id: str  # ID!

    # Optional fields
    client_mutation_id: str | None | UnsetType = Field(
        default=UNSET, alias="clientMutationId"
    )  # String
    title: str | None | UnsetType = UNSET  # String
    code: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = UNSET  # Int (1-100)
    organized: bool | None | UnsetType = UNSET  # Boolean
    urls: list[str] | None | UnsetType = UNSET  # [String!]
    date: str | None | UnsetType = UNSET  # String
    details: str | None | UnsetType = UNSET  # String
    photographer: str | None | UnsetType = UNSET  # String
    studio_id: str | None | UnsetType = UNSET  # ID
    performer_ids: list[str] | None | UnsetType = UNSET  # [ID!]
    tag_ids: list[str] | None | UnsetType = UNSET  # [ID!]
    gallery_ids: list[str] | None | UnsetType = UNSET  # [ID!]
    primary_file_id: str | None | UnsetType = UNSET  # ID


class Image(StashObject):
    """Image type from schema."""

    __type_name__ = "Image"
    __update_input_type__ = ImageUpdateInput
    # No __create_input_type__ - images can only be updated

    # Fields to track for changes - only fields that can be written via input types
    __tracked_fields__ = {
        "title",  # ImageUpdateInput
        "code",  # ImageUpdateInput
        "urls",  # ImageUpdateInput
        "date",  # ImageUpdateInput
        "details",  # ImageUpdateInput
        "photographer",  # ImageUpdateInput
        "studio",  # mapped to studio_id
        "organized",  # ImageUpdateInput
        "galleries",  # mapped to gallery_ids
        "tags",  # mapped to tag_ids
        "performers",  # mapped to performer_ids
    }

    # Optional fields
    title: str | None | UnsetType = UNSET  # String
    code: str | None | UnsetType = UNSET  # String
    date: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = Field(
        default=UNSET, ge=0, le=100
    )  # Int (0-100)
    details: str | None | UnsetType = UNSET  # String
    photographer: str | None | UnsetType = UNSET  # String
    studio: Studio | None | UnsetType = UNSET  # Studio
    o_counter: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int (Resolver)

    # Required fields
    urls: list[str] | UnsetType = Field(default=UNSET)  # [String!]!
    organized: bool | UnsetType = UNSET  # Boolean!
    visual_files: list[VisualFile] | UnsetType = Field(
        default=UNSET
    )  # [VisualFile!]! - The image files (union of VideoFile | ImageFile)
    paths: ImagePathsType | UnsetType = Field(
        default=UNSET
    )  # ImagePathsType! (Resolver)
    galleries: list[Gallery] | UnsetType = Field(default=UNSET)  # [Gallery!]!
    tags: list[Tag] | UnsetType = Field(default=UNSET)  # [Tag!]!
    performers: list[Performer] | UnsetType = Field(default=UNSET)  # [Performer!]!

    # Relationship definitions with their mappings
    __relationships__ = {
        # Pattern B: Filter query relationship (many-to-one)
        "studio": RelationshipMetadata(
            target_field="studio_id",
            is_list=False,
            query_field="studio",
            inverse_type="Studio",
            query_strategy="filter_query",
            filter_query_hint="findImages(image_filter={studios: {value: [studio_id]}})",
            notes="Studio has image_count and filter queries, not direct images field",
        ),
        # Pattern A: Direct field relationships (many-to-many)
        "performers": RelationshipMetadata(
            target_field="performer_ids",
            is_list=True,
            query_field="performers",
            inverse_type="Performer",
            query_strategy="direct_field",
            notes="Performer has image_count resolver, not direct images list",
        ),
        "tags": RelationshipMetadata(
            target_field="tag_ids",
            is_list=True,
            query_field="tags",
            inverse_type="Tag",
            query_strategy="direct_field",
            notes="Tag has image_count resolver, not direct images list",
        ),
        "galleries": RelationshipMetadata(
            target_field="gallery_ids",
            is_list=True,
            query_field="galleries",
            inverse_type="Gallery",
            query_strategy="direct_field",
            notes="Gallery has image_count and image(index) method, not direct images list",
        ),
    }

    # Field definitions with their conversion functions
    __field_conversions__ = {
        "title": str,
        "code": str,
        "urls": list,
        "details": str,
        "photographer": str,
        "rating100": int,
        "organized": bool,
        "date": lambda d: (
            d.strftime("%Y-%m-%d")
            if isinstance(d, datetime)
            else (
                datetime.fromisoformat(d).strftime("%Y-%m-%d")
                if isinstance(d, str)
                else None
            )
        ),
    }

    async def add_performer(self, performer: Performer) -> None:
        """Add performer (syncs inverse automatically, call save() to persist)."""
        await self._add_to_relationship("performers", performer)

    async def remove_performer(self, performer: Performer) -> None:
        """Remove performer (syncs inverse automatically, call save() to persist)."""
        await self._remove_from_relationship("performers", performer)

    async def add_to_gallery(self, gallery: Gallery) -> None:
        """Add gallery (syncs inverse automatically, call save() to persist)."""
        await self._add_to_relationship("galleries", gallery)

    async def remove_from_gallery(self, gallery: Gallery) -> None:
        """Remove gallery (syncs inverse automatically, call save() to persist)."""
        await self._remove_from_relationship("galleries", gallery)

    async def add_tag(self, tag: Tag) -> None:
        """Add tag (syncs inverse automatically, call save() to persist)."""
        await self._add_to_relationship("tags", tag)

    async def remove_tag(self, tag: Tag) -> None:
        """Remove tag (syncs inverse automatically, call save() to persist)."""
        await self._remove_from_relationship("tags", tag)

    @field_validator("visual_files", mode="before")
    @classmethod
    def _discriminate_visual_file_types(cls, value: Any) -> Any:
        """Discriminate VisualFile union types based on __typename.

        Converts raw dicts to the appropriate VisualFile type (VideoFile or ImageFile)
        based on __typename field.
        """
        if value is None or isinstance(value, type(UNSET)):
            return value

        if not isinstance(value, list):
            return value

        # Type mapping for VisualFile union
        type_map: dict[str, type[VideoFile] | type[ImageFile]] = {
            "VideoFile": VideoFile,
            "ImageFile": ImageFile,
        }

        result = []
        for item in value:
            if isinstance(item, dict) and "__typename" in item:
                typename = item["__typename"]
                file_type = type_map.get(typename)
                if file_type:
                    # Construct the correct type from the dict
                    result.append(file_type.model_validate(item))
                else:
                    # Unknown typename, skip
                    continue
            else:
                # Already constructed or no typename
                result.append(item)

        return result


class ImageDestroyInput(StashInput):
    """Input for destroying images from schema/types/image.graphql."""

    id: str  # ID!
    delete_file: bool | None | UnsetType = UNSET  # Boolean
    delete_generated: bool | None | UnsetType = UNSET  # Boolean


class ImagesDestroyInput(StashInput):
    """Input for destroying multiple images from schema/types/image.graphql."""

    ids: list[str] | UnsetType = UNSET  # [ID!]!
    delete_file: bool | None | UnsetType = UNSET  # Boolean
    delete_generated: bool | None | UnsetType = UNSET  # Boolean


class BulkImageUpdateInput(StashInput):
    """Input for bulk updating images."""

    # Optional fields
    client_mutation_id: str | None | UnsetType = Field(
        default=UNSET, alias="clientMutationId"
    )  # String
    ids: list[str] | UnsetType = UNSET  # [ID!]
    title: str | None | UnsetType = UNSET  # String
    code: str | None | UnsetType = UNSET  # String
    rating100: int | None | UnsetType = Field(
        default=UNSET, ge=0, le=100
    )  # Int (0-100)
    organized: bool | None | UnsetType = UNSET  # Boolean
    urls: BulkUpdateStrings | None | UnsetType = UNSET  # BulkUpdateStrings
    date: str | None | UnsetType = UNSET  # String
    details: str | None | UnsetType = UNSET  # String
    photographer: str | None | UnsetType = UNSET  # String
    studio_id: str | None | UnsetType = UNSET  # ID
    performer_ids: BulkUpdateIds | None | UnsetType = UNSET  # BulkUpdateIds
    tag_ids: BulkUpdateIds | None | UnsetType = UNSET  # BulkUpdateIds
    gallery_ids: BulkUpdateIds | None | UnsetType = UNSET  # BulkUpdateIds


class FindImagesResultType(StashResult):
    """Result type for finding images from schema/types/image.graphql."""

    count: int | UnsetType = UNSET  # Int!
    megapixels: float | UnsetType = UNSET  # Float! (Total megapixels of the images)
    filesize: float | UnsetType = UNSET  # Float! (Total file size in bytes)
    images: list[Image] | UnsetType = Field(default=UNSET)  # [Image!]!



"""Statistics types from schema/types/stats.graphql."""

from __future__ import annotations

from pydantic import Field

from .base import StashResult
from .unset import UNSET, UnsetType


class StatsResultType(StashResult):
    """Statistics result type from schema/types/stats.graphql."""

    scene_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    scenes_size: float | None | UnsetType = UNSET  # Float!
    scenes_duration: float | None | UnsetType = UNSET  # Float!
    image_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    images_size: float | None | UnsetType = UNSET  # Float!
    gallery_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    performer_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    studio_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    group_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    tag_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    total_o_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    total_play_duration: float | None | UnsetType = UNSET  # Float!
    total_play_count: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
    scenes_played: int | None | UnsetType = Field(default=UNSET, ge=0)  # Int!
